@@ -107,7 +107,7 @@ static char *if_names[] = { "Auto", "10baseT", "10base2", "AUI" };
 #ifdef PCMCIA_DEBUG
 static int pc_debug = PCMCIA_DEBUG;
 static char *version =
-"3c589_cs.c 1.81 1996/06/03 17:07:03 (David Hinds)";
+"3c589_cs.c 1.84 1996/08/16 05:01:58 (David Hinds)";
 #endif
 
 #ifdef TC589_DEBUG
@@ -343,7 +343,7 @@ static void tc589_config(dev_link_t *link)
 	if (i != CS_SUCCESS) break;
 	i = CardServices(ParseTuple, handle, &tuple, &parse);
 	if (i != CS_SUCCESS) break;
-	link->conf.ConfigBase = (caddr_t)parse.config.base;
+	link->conf.ConfigBase = parse.config.base;
 	link->conf.Present = parse.config.rmask[0];
     } while (0);
     if (i != CS_SUCCESS) {
@@ -628,8 +628,8 @@ static void tc589_reset(struct device *dev)
 
 static int el3_config(struct device *dev, struct ifmap *map)
 {
-    if (map->port != (u_char)(-1)) {
-	if ((map->port >= 1) || (map->port <= 3)) {
+    if ((map->port != (u_char)(-1)) && (map->port != dev->if_port)) {
+	if ((map->port >= 1) && (map->port <= 3)) {
 	    dev->if_port = map->port;
 	    printk(KERN_INFO "%s: switched to %s port\n",
 		   dev->name, if_names[dev->if_port]);
@@ -1038,11 +1038,11 @@ static void set_multicast_list(struct device *dev)
 		   dev->name, old);
 	}
     }
-    if (dev->mc_count || (dev->flags & IFF_ALLMULTI))
-	outw(SetRxFilter|RxStation|RxMulticast|RxBroadcast, ioaddr + EL3_CMD);
-    else if (dev->flags & IFF_PROMISC)
+    if (dev->flags & IFF_PROMISC)
 	outw(SetRxFilter | RxStation | RxMulticast | RxBroadcast | RxProm,
 	     ioaddr + EL3_CMD);
+    else if (dev->mc_count || (dev->flags & IFF_ALLMULTI))
+	outw(SetRxFilter|RxStation|RxMulticast|RxBroadcast, ioaddr + EL3_CMD);
     else
 	outw(SetRxFilter | RxStation | RxBroadcast, ioaddr + EL3_CMD);
 }
