@@ -6,11 +6,9 @@
     
 ======================================================================*/
 
-#ifdef MODULE
-#include <linux/autoconf.h>
-#include <linux/module.h>
-#include <linux/version.h>
-#endif
+#include "config.h"
+#define __NO_VERSION__
+#include "k_compat.h"
 
 #include <linux/kernel.h>
 #include <linux/string.h>
@@ -20,11 +18,12 @@
 #include <linux/malloc.h>
 #include <linux/mm.h>
 #include <linux/sched.h>
+#include <asm/io.h>
 
 #include "cs_types.h"
-#include "rsrc_mgr.h"
 #include "ss.h"
 #include "cs.h"
+#include "rsrc_mgr.h"
 #include "bulkmem.h"
 #include "cs_internal.h"
 #include "cistpl.h"
@@ -66,7 +65,8 @@ void read_cis_mem(socket_info_t *s, int attr,
     
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 3)
-	printk("cs: read_cis_mem(%d, %#lx, %ld)\n", attr, addr, len);
+	printk(KERN_DEBUG "cs: read_cis_mem(%d, %#lx, %ld)\n",
+	       attr, addr, len);
 #endif
     mem->card_start = addr & ~0x0fff;
     mem->flags &= ~MAP_ATTRIB;
@@ -76,12 +76,12 @@ void read_cis_mem(socket_info_t *s, int attr,
 	s->ss_entry(s->sock, SS_SetMemMap, mem);
 #ifdef PCMCIA_DEBUG
 	if (pc_debug > 3)
-	    printk("cs:  %#2.2x %#2.2x %#2.2x %#2.2x %#2.2x ...\n",
+	    printk(KERN_DEBUG "cs:  %#2.2x %#2.2x %#2.2x %#2.2x %#2.2x ...\n",
 		   sys[0], sys[inc], sys[2*inc], sys[3*inc], sys[4*inc]);
 #endif
 	for ( ; len > 0; len--, ptr++, sys += inc) {
 	    if (sys > (u_char *)mem->sys_stop) break;
-	    *ptr = *sys;
+	    *ptr = readb(sys);
 	}
     }
 } /* read_cis_mem */
@@ -101,7 +101,7 @@ void write_cis_mem(socket_info_t *s, int attr,
 	s->ss_entry(s->sock, SS_SetMemMap, mem);
 	for ( ; len > 0; len--, ptr++, sys += inc) {
 	    if (sys > (u_char *)mem->sys_stop) break;
-	    *sys = *ptr;
+	    writeb(*ptr, sys);
 	}
     }
 } /* write_cis_mem */
@@ -117,24 +117,94 @@ int setup_cis_mem(socket_info_t *s)
     return 0;
 } /* setup_cis_mem */
 
+/*======================================================================
+
+    This is a wrapper around read_cis_mem, with the same interface,
+    but which caches information, for cards whose CIS may not be
+    readable all the time.
+    
+======================================================================*/
+
+static void read_cis_cache(socket_info_t *s, int attr,
+			   long addr, long len, u_char *ptr)
+{
+    int i;
+    char *caddr;
+
+    caddr = s->cis_cache;
+    for (i = 0; i < s->cis_used; i++) {
+	if ((s->cis_table[i].addr == addr) &&
+	    (s->cis_table[i].len == len) &&
+	    (s->cis_table[i].attr == attr)) break;
+	caddr += s->cis_table[i].len;
+    }
+    if (i < s->cis_used) {
+	memcpy(ptr, caddr, len);
+	return;
+    }
+    read_cis_mem(s, attr, addr, len, ptr);
+    /* Copy data into the cache, if there is room */
+    if ((i < MAX_CIS_TABLE) &&
+	(caddr+len < s->cis_cache+MAX_CIS_DATA)) {
+	s->cis_table[i].addr = addr;
+	s->cis_table[i].len = len;
+	s->cis_table[i].attr = attr;
+	s->cis_used++;
+	memcpy(caddr, ptr, len);
+    }	    
+}
+
+/*======================================================================
+
+    This verifies if the CIS of a card matches what is in the CIS
+    cache.
+    
+======================================================================*/
+
+int verify_cis_cache(socket_info_t *s)
+{
+    char buf[256], *caddr;
+    int i;
+    
+    caddr = s->cis_cache;
+    for (i = 0; i < s->cis_used; i++) {
+	read_cis_mem(s, s->cis_table[i].attr, s->cis_table[i].addr,
+		     s->cis_table[i].len, buf);
+	if (memcmp(buf, caddr, s->cis_table[i].len) != 0)
+	    break;
+	caddr += s->cis_table[i].len;
+    }
+    return (i < s->cis_used);
+}
+
 /*======================================================================
 
     Read tuple codes from attribute memory.
     
 ======================================================================*/
 
-#define LINK_C	0x01
-#define LINK_A	0x02
-#define IS_ATTR 0x04
+#define LINK_MFC_N	0x00ff
+#define LINK_C		0x0100
+#define LINK_A		0x0200
+#define LINK_MFC	0x0400
+#define IS_ATTR 	0x1000
+#define MFC_IS_ATTR	0x2000
 
 int get_next_tuple(client_handle_t handle, tuple_t *tuple);
 
 int get_first_tuple(client_handle_t handle, tuple_t *tuple)
 {
+    if (CHECK_HANDLE(handle))
+	return CS_BAD_HANDLE;
     tuple->CISOffset = -1;
-    /* Assume presence of a LONGLINK_C to address 0 */
-    tuple->LinkOffset = 0;
-    tuple->Flags = IS_ATTR | LINK_C;
+    if ((SOCKET(handle)->functions > 1) &&
+	!(tuple->Attributes & TUPLE_RETURN_COMMON)) {
+	cisdata_t req = tuple->DesiredTuple;
+	tuple->DesiredTuple = CISTPL_LONGLINK_MFC;
+	if (get_next_tuple(handle, tuple) != CS_SUCCESS)
+	    tuple->CISOffset = -1;
+	tuple->DesiredTuple = req;
+    }
     return get_next_tuple(handle, tuple);
 } /* get_first_tuple */
 
@@ -143,7 +213,7 @@ static long follow_link(socket_info_t *s, tuple_t *tuple)
     u_char link[5];
     long ofs;
     
-    switch (tuple->Flags & (LINK_A|LINK_C)) {
+    switch (tuple->Flags & (LINK_A|LINK_C|LINK_MFC)) {
     case LINK_A:
 	ofs = tuple->LinkOffset;
 	tuple->Flags = IS_ATTR;
@@ -152,10 +222,36 @@ static long follow_link(socket_info_t *s, tuple_t *tuple)
 	ofs = tuple->LinkOffset;
 	tuple->Flags = 0;
 	break;
+    case LINK_MFC:
+	/* Get indirect link from the MFC tuple */
+	read_cis_cache(s, tuple->Flags & MFC_IS_ATTR,
+		       tuple->LinkOffset, 5, link);
+	ofs = *(long *)(link+1);
+	if (link[0] == CISTPL_MFC_ATTR)
+	    tuple->Flags |= IS_ATTR;
+	else
+	    tuple->Flags &= ~IS_ATTR;
+	/* Move to the next indirect link */
+	tuple->Flags--;
+	tuple->LinkOffset += (tuple->Flags & MFC_IS_ATTR) ? 10 : 5;
+	if ((tuple->Flags & LINK_MFC_N) == 0)
+	    tuple->Flags &= ~LINK_MFC;
+	break;
     default:
 	return -1;
     }
-    read_cis_mem(s, tuple->Flags, ofs, 5, link);
+    if (tuple->Flags & IS_ATTR) {
+	ofs = ofs << 1;
+	/* This is ugly, but a common CIS error is to code the long
+	   link offset incorrectly, so we check the right spot... */
+	read_cis_cache(s, tuple->Flags & IS_ATTR, ofs, 5, link);
+	if ((link[0] == CISTPL_LINKTARGET) && (link[1] >= 3) &&
+	    (strncmp(link+2, "CIS", 3) == 0))
+	    return ofs;
+	/* Then, we try the wrong spot... */
+	ofs = ofs >> 1;
+    }
+    read_cis_cache(s, tuple->Flags & IS_ATTR, ofs, 5, link);
     if ((link[0] != CISTPL_LINKTARGET) || (link[1] < 3) ||
 	(strncmp(link+2, "CIS", 3) != 0))
 	return -1;
@@ -165,9 +261,9 @@ static long follow_link(socket_info_t *s, tuple_t *tuple)
 int get_next_tuple(client_handle_t handle, tuple_t *tuple)
 {
     socket_info_t *s;
-    u_char link[2], attr;
+    u_char link[2], tmp;
     long ofs;
-    int i;
+    int i, attr;
     
     if (CHECK_HANDLE(handle))
 	return CS_BAD_HANDLE;
@@ -176,21 +272,25 @@ int get_next_tuple(client_handle_t handle, tuple_t *tuple)
     if (setup_cis_mem(s) != 0)
 	return CS_OUT_OF_RESOURCE;
 
-    attr = tuple->Flags & IS_ATTR;
-    if (tuple->CISOffset == -1)
+    if (tuple->CISOffset == -1) {
 	ofs = 0;
+	/* Assume presence of a LONGLINK_C to address 0 */
+	tuple->LinkOffset = 0;
+	tuple->Flags = IS_ATTR | LINK_C;
+    }
     else {
 	if (tuple->TupleLink == 0xff) {
 	    if ((ofs = follow_link(s, tuple)) < 0)
 		return CS_NO_MORE_ITEMS;
-	    attr = tuple->Flags & IS_ATTR;
 	}
 	else
-	    ofs = tuple->CISOffset + (attr ? 2 : 1) * (tuple->TupleLink+2);
+	    ofs = tuple->CISOffset +
+		((tuple->Flags & IS_ATTR) ? 2 : 1) * (tuple->TupleLink+2);
     }
+    attr = tuple->Flags & IS_ATTR;
     
     for (i = 0; i < 1000; i++) {
-	read_cis_mem(s, attr, ofs, 2, link);
+	read_cis_cache(s, attr, ofs, 2, link);
 
 	if (link[0] == CISTPL_NULL) {
 	    ofs += attr ? 2 : 1;
@@ -202,24 +302,42 @@ int get_next_tuple(client_handle_t handle, tuple_t *tuple)
 	    if ((ofs = follow_link(s, tuple)) < 0)
 		return CS_NO_MORE_ITEMS;
 	    attr = tuple->Flags & IS_ATTR;
-	    read_cis_mem(s, attr, ofs, 2, link);
+	    read_cis_cache(s, attr, ofs, 2, link);
 	}
 
 	/* Is this a link tuple?  Make a note of it */
 	if ((link[0] == CISTPL_LONGLINK_A) ||
 	    (link[0] == CISTPL_LONGLINK_C) ||
+	    (link[0] == CISTPL_LONGLINK_MFC) ||
 	    (link[0] == CISTPL_LINKTARGET) ||
 	    (link[0] == CISTPL_NO_LINK)) {
 	    switch (link[0]) {
 	    case CISTPL_LONGLINK_A:
 		tuple->Flags = attr | LINK_A;
-		read_cis_mem(s, attr, ofs + (attr ? 4 : 2), 4,
-			     (u_char *)&tuple->LinkOffset);
+		read_cis_cache(s, attr, ofs + (attr ? 4 : 2), 4,
+			       (u_char *)&tuple->LinkOffset);
 		break;
 	    case CISTPL_LONGLINK_C:
 		tuple->Flags = attr | LINK_C;
-		read_cis_mem(s, attr, ofs + (attr ? 4 : 2), 4,
-			     (u_char *)&tuple->LinkOffset);
+		read_cis_cache(s, attr, ofs + (attr ? 4 : 2), 4,
+			       (u_char *)&tuple->LinkOffset);
+		break;
+	    case CISTPL_LONGLINK_MFC:
+		if (handle->Function == BIND_FN_ALL) {
+		    /* Follow all the MFC links */
+		    read_cis_cache(s, attr, ofs + (attr ? 4 : 2), 1,
+				 (u_char *)&tmp);
+		    tuple->Flags = attr | LINK_MFC | tmp;
+		    tuple->LinkOffset = ofs + (attr ? 6 : 3);
+		}
+		else {
+		    /* Follow exactly one of the links */
+		    tuple->Flags = attr | LINK_MFC | 1;
+		    tuple->LinkOffset = ofs + (attr ? 2 : 1) *
+			(3 + handle->Function * 5);
+		}
+		if (attr)
+		    tuple->Flags |= MFC_IS_ATTR;
 		break;
 	    case CISTPL_NO_LINK:
 		tuple->Flags = attr;
@@ -238,8 +356,8 @@ int get_next_tuple(client_handle_t handle, tuple_t *tuple)
 	ofs += (attr ? 2 : 1) * (link[1]+2);
     }
     if (i == 1000) {
-	printk("cs: overrun in get_next_tuple for socket %d\n",
-	       handle->Socket);
+	printk(KERN_NOTICE "cs: overrun in get_next_tuple for "
+	       "socket %d\n", handle->Socket);
 	return CS_NO_MORE_ITEMS;
     }
     
@@ -270,11 +388,11 @@ int get_tuple_data(client_handle_t handle, tuple_t *tuple)
     if (len == 0)
 	return CS_SUCCESS;
     if (tuple->Flags & IS_ATTR)
-	read_cis_mem(s, 1, tuple->CISOffset+((tuple->TupleOffset+2)<<1),
-		     MIN(len, tuple->TupleDataMax), tuple->TupleData);
+	read_cis_cache(s, 1, tuple->CISOffset+((tuple->TupleOffset+2)<<1),
+		       MIN(len, tuple->TupleDataMax), tuple->TupleData);
     else
-	read_cis_mem(s, 0, tuple->CISOffset+(tuple->TupleOffset+2),
-		     MIN(len, tuple->TupleDataMax), tuple->TupleData);
+	read_cis_cache(s, 0, tuple->CISOffset+(tuple->TupleOffset+2),
+		       MIN(len, tuple->TupleDataMax), tuple->TupleData);
     return CS_SUCCESS;
 } /* get_tuple_data */
 
@@ -304,7 +422,8 @@ int validate_cis(client_handle_t handle, cisinfo_t *info)
 	if (ret != CS_SUCCESS) break;
     }
     if (info->Chains == 500) {
-	printk("cs: too many tuples for socket %d\n", handle->Socket);
+	printk(KERN_NOTICE "cs: too many tuples for socket %d\n",
+	       handle->Socket);
 	info->Chains = 0;
     }
     
@@ -388,6 +507,26 @@ static int parse_longlink(tuple_t *tuple, cistpl_longlink_t *link)
 
 /*====================================================================*/
 
+static int parse_longlink_mfc(tuple_t *tuple,
+			      cistpl_longlink_mfc_t *link)
+{
+    u_char *p;
+    int i;
+    
+    p = (char *)tuple->TupleData;
+    
+    link->nfn = *p; p++;
+    if (tuple->TupleDataLen <= link->nfn*5)
+	return CS_BAD_TUPLE;
+    for (i = 0; i < link->nfn; i++) {
+	link->fn[i].space = *p; p++;
+	link->fn[i].addr = *(u_long *)p; p += 4;
+    }
+    return CS_SUCCESS;
+}
+
+/*====================================================================*/
+
 static int parse_strings(u_char *p, u_char *q, int max,
 			 char *s, int *ofs, int *found)
 {
@@ -496,6 +635,21 @@ static int parse_funcid(tuple_t *tuple, cistpl_funcid_t *f)
 
 /*====================================================================*/
 
+static int parse_funce(tuple_t *tuple, cistpl_funce_t *f)
+{
+    u_char *p;
+    int i;
+    if (tuple->TupleDataLen < 1)
+	return CS_BAD_TUPLE;
+    p = (u_char *)tuple->TupleData;
+    f->type = p[0];
+    for (i = 1; i < tuple->TupleDataLen; i++)
+	f->data[i-1] = p[i];
+    return CS_SUCCESS;
+} /* parse_funce */
+
+/*====================================================================*/
+
 static int parse_config(tuple_t *tuple, cistpl_config_t *config)
 {
     int rasz, rmsz, i;
@@ -601,7 +755,9 @@ static u_char *parse_io(u_char *p, u_char *q, cistpl_io_t *io)
     io->flags = *p;
 
     if (!(*p & 0x80)) {
-	io->nwin = 0;
+	io->nwin = 1;
+	io->win[0].base = 0;
+	io->win[0].len = (1 << (io->flags & CISTPL_IO_LINES_MASK));
 	return p+1;
     }
     
@@ -886,6 +1042,9 @@ int parse_tuple(client_handle_t handle, tuple_t *tuple, cisparse_t *parse)
     case CISTPL_LONGLINK_C:
 	ret = parse_longlink(tuple, &parse->longlink);
 	break;
+    case CISTPL_LONGLINK_MFC:
+	ret = parse_longlink_mfc(tuple, &parse->longlink_mfc);
+	break;
     case CISTPL_VERS_1:
 	ret = parse_vers_1(tuple, &parse->version_1);
 	break;
@@ -902,6 +1061,9 @@ int parse_tuple(client_handle_t handle, tuple_t *tuple, cisparse_t *parse)
     case CISTPL_FUNCID:
 	ret = parse_funcid(tuple, &parse->funcid);
 	break;
+    case CISTPL_FUNCE:
+	ret = parse_funce(tuple, &parse->funce);
+	break;
     case CISTPL_CONFIG:
 	ret = parse_config(tuple, &parse->config);
 	break;
@@ -924,3 +1086,28 @@ int parse_tuple(client_handle_t handle, tuple_t *tuple, cisparse_t *parse)
     }
     return ret;
 } /* parse_tuple */
+
+/*======================================================================
+
+    This is used internally by Card Services to look up CIS stuff.
+    
+======================================================================*/
+
+int read_tuple(client_handle_t handle, cisdata_t code, void *parse)
+{
+    tuple_t tuple;
+    cisdata_t buf[255];
+    int ret;
+    
+    tuple.DesiredTuple = code;
+    tuple.Attributes = 0;
+    ret = CardServices(GetFirstTuple, handle, &tuple, NULL);
+    if (ret != CS_SUCCESS) return ret;
+    tuple.TupleData = buf;
+    tuple.TupleOffset = 0;
+    tuple.TupleDataMax = sizeof(buf);
+    ret = CardServices(GetTupleData, handle, &tuple, NULL);
+    if (ret != CS_SUCCESS) return ret;
+    ret = CardServices(ParseTuple, handle, &tuple, parse);
+    return ret;
+}
