@@ -45,7 +45,7 @@ static int handle_apm_event(apm_event_t event);
 #ifdef PCMCIA_DEBUG
 int pc_debug = PCMCIA_DEBUG;
 static const char *version =
-"cs.c 1.132 1996/06/04 16:26:16 (David Hinds)";
+"cs.c 1.138 1996/08/06 03:11:49 (David Hinds)";
 #endif
 
 static const char *release =
@@ -70,7 +70,7 @@ static struct symbol_table cs_symtab = {
 /* Parameters that can be set with 'insmod' */
 
 static int setup_delay		= HZ/20;	/* ticks */
-static int shutdown_delay	= HZ/20;	/* ticks */
+static int shutdown_delay	= HZ/40;	/* ticks */
 static int vcc_settle		= HZ*0.3;	/* ticks */
 static int reset_time		= 10;		/* usecs */
 static int unreset_check	= HZ/10;	/* ticks */
@@ -344,6 +344,7 @@ static void shutdown_socket(u_long i)
     s->irq.AssignedIRQ = s->irq.Config = 0;
     s->functions = 0;
     s->lock_count = 0;
+    s->cis_used = 0;
     if (s->config) {
 	kfree(s->config);
 	s->config = NULL;
@@ -423,7 +424,10 @@ static void unreset_socket(u_long i)
 	    s->state &= ~SOCKET_SETUP_PENDING;
 	}
 	else if (s->state & SOCKET_SUSPEND) {
-	    send_event(s, CS_EVENT_PM_RESUME, CS_EVENT_PRI_LOW);
+	    if (verify_cis_cache(s) != 0)
+		parse_events(s, SS_DETECT);
+	    else
+		send_event(s, CS_EVENT_PM_RESUME, CS_EVENT_PRI_LOW);
 	    s->state &= ~SOCKET_SUSPEND;
 	}
 	else {
@@ -573,7 +577,8 @@ static int handle_apm_event(apm_event_t event)
 	down = 1;
 	for (i = 0; i < sockets; i++) {
 	    s = socket_table[i];
-	    if (s->state & SOCKET_PRESENT) {
+	    if ((s->state & SOCKET_PRESENT) &&
+		!(s->state & SOCKET_SUSPEND)){
 		send_event(s, CS_EVENT_PM_SUSPEND, CS_EVENT_PRI_LOW);
 		s->ss_entry(s->sock, SS_SetSocket, &dead_socket);
 		s->state |= SOCKET_SUSPEND;
@@ -604,8 +609,7 @@ static int handle_apm_event(apm_event_t event)
 		else
 		    /* We missed a removal event */
 		    parse_events(s, SS_DETECT);
-	    }
-	    else {
+	    } else {
 		if (stat & SS_DETECT)
 		    /* we missed an insertion event */
 		    parse_events(s, SS_DETECT);
@@ -712,7 +716,7 @@ static int access_configuration_register(client_handle_t handle,
 
     if (setup_cis_mem(s) != 0)
 	return CS_OUT_OF_RESOURCE;
-    addr = (long)c->ConfigBase + reg->Offset;
+    addr = c->ConfigBase + reg->Offset;
     
     switch (reg->Action) {
     case CS_READ:
@@ -908,7 +912,7 @@ static int get_configuration_info(client_handle_t handle,
 	memcpy(&config->Attributes, &c->Attributes, 33);
 	config->Attributes |= CONF_VALID_CLIENT;
 	config->CardValues = c->CardValues;
-	config->IRQAttributes = s->irq.Attributes;
+	config->IRQAttributes = c->irq.Attributes;
 	config->AssignedIRQ = s->irq.AssignedIRQ;
 	config->BasePort1 = c->io.BasePort1;
 	config->NumPorts1 = c->io.NumPorts1;
@@ -1005,7 +1009,7 @@ static int get_status(client_handle_t handle, status_t *status)
     if ((c->state & CONFIG_LOCKED) && (c->IntType & INT_MEMORY_AND_IO)) {
 	u_char reg;
 	if (c->Present & PRESENT_PIN_REPLACE) {
-	    read_cis_mem(s, 1, (long)c->ConfigBase+CISREG_PRR, 1, &reg);
+	    read_cis_mem(s, 1, c->ConfigBase+CISREG_PRR, 1, &reg);
 	    status->CardState |=
 		(reg & PRR_WP_STATUS) ? CS_EVENT_WRITE_PROTECT : 0;
 	    status->CardState |=
@@ -1016,7 +1020,7 @@ static int get_status(client_handle_t handle, status_t *status)
 		(reg & PRR_BVD1_STATUS) ? CS_EVENT_BATTERY_DEAD : 0;
 	}
 	if (c->Present & PRESENT_EXT_STATUS) {
-	    read_cis_mem(s, 1, (long)c->ConfigBase+CISREG_ESR, 1, &reg);
+	    read_cis_mem(s, 1, c->ConfigBase+CISREG_ESR, 1, &reg);
 	    status->CardState |=
 		(reg & ESR_REQ_ATTN) ? CS_EVENT_REQUEST_ATTENTION : 0;
 	}
@@ -1221,9 +1225,8 @@ static int release_configuration(client_handle_t handle,
     if (!(handle->state & CLIENT_STALE)) {
 	socket_info_t *s = SOCKET(handle);
 	config_t *c = CONFIG(handle);
-	if (!(c->state & CLIENT_CONFIG_LOCKED))
+	if (!(c->state & CONFIG_LOCKED))
 	    return CS_BAD_HANDLE;
-	c->state &= ~CONFIG_LOCKED;
 	if (--(s->lock_count) == 0) {
 	    s->socket.flags = SS_OUTPUT_ENA;
 	    s->socket.Vcc = 50; s->socket.Vpp = 0;
@@ -1299,7 +1302,7 @@ static int release_irq(client_handle_t handle, irq_req_t *req)
 	config_t *c = CONFIG(handle);
 	if (c->state & CONFIG_LOCKED)
 	    return CS_CONFIGURATION_LOCKED;
-	if (s->irq.Attributes != req->Attributes)
+	if (c->irq.Attributes != req->Attributes)
 	    return CS_BAD_ATTRIBUTE;
 	if (s->irq.AssignedIRQ != req->AssignedIRQ)
 	    return CS_BAD_IRQ;
@@ -1400,8 +1403,7 @@ static int request_configuration(client_handle_t handle,
     c->Present = c->CardValues = req->Present;
     if (req->Present & PRESENT_COPY) {
 	c->Copy = req->Copy;
-	write_cis_mem(s, 1, (long)req->ConfigBase+CISREG_SCR,
-		      1, &c->Copy);
+	write_cis_mem(s, 1, req->ConfigBase+CISREG_SCR, 1, &c->Copy);
     }
     if (req->Present & PRESENT_OPTION) {
 	if (s->functions == 1)
@@ -1411,38 +1413,34 @@ static int request_configuration(client_handle_t handle,
 	    c->Option |= COR_FUNC_ENA|COR_ADDR_DECODE|COR_IREQ_ENA;
 	}
 	if (c->state & CONFIG_IRQ_REQ)
-	    if (!(s->irq.Attributes & IRQ_FORCED_PULSE))
+	    if (!(c->irq.Attributes & IRQ_FORCED_PULSE))
 		c->Option |= COR_LEVEL_REQ;
-	write_cis_mem(s, 1, (long)req->ConfigBase+CISREG_COR,
-		      1, &c->Option);
+	write_cis_mem(s, 1, req->ConfigBase+CISREG_COR, 1, &c->Option);
 	busy_loop(HZ/25);
     }
     if (req->Present & PRESENT_STATUS) {
 	c->Status = req->Status;
-	write_cis_mem(s, 1, (long)req->ConfigBase+CISREG_CCSR,
-		      1, &c->Status);
+	write_cis_mem(s, 1, req->ConfigBase+CISREG_CCSR, 1, &c->Status);
     }
     if (req->Present & PRESENT_PIN_REPLACE) {
 	c->Pin = req->Pin;
-	write_cis_mem(s, 1, (long)req->ConfigBase+CISREG_PRR,
-		      1, &c->Pin);
+	write_cis_mem(s, 1, req->ConfigBase+CISREG_PRR, 1, &c->Pin);
     }
     if (req->Present & PRESENT_EXT_STATUS) {
 	c->ExtStatus = req->ExtStatus;
-	write_cis_mem(s, 1, (long)req->ConfigBase+CISREG_ESR,
-		      1, &c->ExtStatus);
+	write_cis_mem(s, 1, req->ConfigBase+CISREG_ESR, 1, &c->ExtStatus);
     }
     if (req->Present & PRESENT_IOBASE_0) {
 	i = c->io.BasePort1 & 0xff;
-	write_cis_mem(s, 1, (long)req->ConfigBase+CISREG_IOBASE_0,
+	write_cis_mem(s, 1, req->ConfigBase+CISREG_IOBASE_0,
 		      1, (caddr_t)&i);
 	i = (c->io.BasePort1 >> 8) & 0xff;
-	write_cis_mem(s, 1, (long)req->ConfigBase+CISREG_IOBASE_1,
+	write_cis_mem(s, 1, req->ConfigBase+CISREG_IOBASE_1,
 		      1, (caddr_t)&i);
     }
     if (req->Present & PRESENT_IOSIZE) {
 	i = c->io.NumPorts1 + c->io.NumPorts2 - 1;
-	write_cis_mem(s, 1, (long)req->ConfigBase+CISREG_IOSIZE,
+	write_cis_mem(s, 1, req->ConfigBase+CISREG_IOSIZE,
 		      1, (caddr_t)&i);
     }
     
@@ -1593,8 +1591,8 @@ static int cs_request_irq(client_handle_t handle, irq_req_t *req)
 #endif
 	return CS_IN_USE;
 
+    c->irq.Attributes = req->Attributes;
     s->irq.AssignedIRQ = req->AssignedIRQ = irq;
-    s->irq.Attributes = req->Attributes;
     s->irq.Config++;
     
     c->state |= CONFIG_IRQ_REQ;
