@@ -1,5 +1,5 @@
 /*
- *  cs_internal.h 1.5 1996/03/10 18:25:08 (David Hinds)
+ *  cs_internal.h 1.8 1996/08/06 03:11:28 (David Hinds)
  */
 
 #ifndef _LINUX_CS_INTERNAL_H
@@ -81,11 +81,14 @@ typedef struct config_t {
     u_long		Attributes;
     u_long		Vcc, Vpp1, Vpp2;
     u_long		IntType;
-    caddr_t		ConfigBase;
+    u_long		ConfigBase;
     u_char		Status, Pin, Copy, Option, ExtStatus;
     u_long		Present;
     u_long		CardValues;
     io_req_t		io;
+    struct {
+	u_long		Attributes;
+    } irq;
 } config_t;
 
 /* Maximum number of IO windows per socket */
@@ -94,6 +97,10 @@ typedef struct config_t {
 /* Maximum number of memory windows per socket */
 #define MAX_WIN 4
 
+/* The size of the CIS cache */
+#define MAX_CIS_TABLE	64
+#define MAX_CIS_DATA	512
+
 typedef struct socket_info_t {
     ss_entry_t		ss_entry;
     u_int		sock;
@@ -109,7 +116,6 @@ typedef struct socket_info_t {
     pcmcia_mem_map	cis_mem;
     config_t		*config;
     struct {
-	u_long		Attributes;
 	u_long		AssignedIRQ;
 	u_long		Config;
     } irq;
@@ -118,6 +124,13 @@ typedef struct socket_info_t {
     region_t		*c_region, *a_region;
     struct wait_queue	*mtd_ready;
     erase_busy_t	erase_busy;
+    int			cis_used;
+    struct {
+	long		addr;
+	u_short		len;
+	u_short		attr;
+    }			cis_table[MAX_CIS_TABLE];
+    char		cis_cache[MAX_CIS_DATA];
 } socket_info_t;
 
 /* Flags in config state */
