@@ -1,10 +1,12 @@
 /*======================================================================
-    fmvj18x_cs.c,v 1.5 1996/05/28 15:17:55 root Exp
+    fmvj18x_cs.c,v 1.9 1996/08/06 03:13:53 root Exp
 
-    A fmvj18x PCMCIA client driver
+    A fmvj18x (and its compatibles) PCMCIA client driver
 
     Written by Shingo Fujimoto, shingo@flab.fujitsu.co.jp
-    
+
+    TDK LAK-CD021 support added by Nobuhiro Katayama, kata-n@po.iijnet.or.jp
+
     Network code is based on the "FMV-18x driver" by Yutaka TAMIYA
 
     This software may be used and distributed according to the terms
@@ -27,6 +29,7 @@
 #include <linux/timer.h>
 #include <linux/interrupt.h>
 #include <linux/in.h>
+#include <linux/delay.h>
 #include <asm/io.h>
 #include <asm/system.h>
 
@@ -72,7 +75,7 @@ static u_long irq_mask = 0xdeb8;
    driver version infomation 
  */
 static char *version =
- "fmvj18x_cs.c,v 1.5 1996/05/28 15:17:55 root Exp";
+ "fmvj18x_cs.c,v 1.9 1996/08/06 03:13:53 root Exp";
 
 /*====================================================================*/
 /*
@@ -107,6 +110,11 @@ static void set_multicast_list(struct device *dev, int num_addrs, void *addrs);
 static dev_info_t dev_info = "fmvj18x_cs";
 static dev_link_t *dev_list = NULL;
 
+/*
+    card type
+ */
+typedef enum { MBH10302, MBH10304, TDK} cardtype_t;
+
 /*
     driver specific data structure
 */
@@ -117,12 +125,10 @@ typedef struct local_info_t {
     uint tx_started:1;
     uint tx_queue;
     ushort tx_queue_len;
+    cardtype_t cardtype;
+    ushort sent;
 } local_info_t;
 
-/*
-    card type
- */
-typedef enum { MBH10302, MBH10304} cardtype_t;
 
 /*====================================================================*/
 /* 
@@ -209,6 +215,8 @@ typedef enum { MBH10302, MBH10304} cardtype_t;
 #define SEND_PKT             0x81 /* send a packet */
 #define AUTO_MODE            0x07 /* Auto skip packet on 16 col detected */
 #define MANU_MODE            0x03 /* Stop and skip packet on 16 col */
+#define TDK_AUTO_MODE        0x47 /* Auto skip packet on 16 col detected */
+#define TDK_MANU_MODE        0x43 /* Stop and skip packet on 16 col */
 #define INTR_OFF             0x0d /* LAN controler ignores interrupts */
 #define INTR_ON              0x1d /* LAN controler will catch interrupts */
 
@@ -255,9 +263,6 @@ static dev_link_t *fmvj18x_attach(void)
     link->conf.Attributes = CONF_ENABLE_IRQ;
     link->conf.Vcc = 50;
     link->conf.IntType = INT_MEMORY_AND_IO;
-#if 0
-    link->conf.ConfigIndex = 1; 
-#endif 
 
     /* Make up a FMVJ18x specific data structure */
     dev = kmalloc(sizeof(struct device), GFP_KERNEL);
@@ -389,14 +394,33 @@ static void fmvj18x_config(dev_link_t *link)
     /* Configure card */
     link->state |= DEV_CONFIG;
 
-    link->conf.ConfigBase = (caddr_t)parse.config.base; 
+    link->conf.ConfigBase = parse.config.base; 
     link->conf.Present = parse.config.rmask[0];
 
     tuple.DesiredTuple = CISTPL_FUNCE;
+    tuple.TupleOffset = 0;
     if (CardServices(GetFirstTuple, handle, &tuple) == CS_SUCCESS) {
-	/* FMV-J182(MBH10304) have MACID on CIS */
-	cardtype = MBH10304;
-	link->conf.ConfigIndex = 0x60;
+	/* Yes, I have CISTPL_FUNCE. Let's check CISTPL_MANFID */
+	tuple.DesiredTuple = CISTPL_MANFID ;
+	tuple.TupleOffset = 0;
+	if (CardServices(GetFirstTuple, handle, &tuple) != CS_SUCCESS) {
+	    fmvj18x_release((u_long)link);
+	    return;
+	}
+	tuple.TupleOffset = 0;
+	if (CardServices(GetTupleData, handle, &tuple) != CS_SUCCESS) {
+	    fmvj18x_release((u_long)link);
+	    return;
+	}
+	if( tuple.TupleData[0] == 0x05 && tuple.TupleData[1] == 0x01 ) {
+	    /* Manufacture is TDK */
+	    cardtype = TDK;
+	    link->conf.ConfigIndex = 1;
+	}else {
+	    /* FMV-J182(MBH10304) have MACID on CIS */
+	    cardtype = MBH10304;
+	    link->conf.ConfigIndex = 0x60;
+	}
     } else {
 	/* old type card */
 	cardtype = MBH10302;
@@ -406,17 +430,17 @@ static void fmvj18x_config(dev_link_t *link)
 	ret = CardServices(RequestIO, link->handle, &link->io);
 	if (ret != CS_SUCCESS) {
 	    cs_error(RequestIO, ret);
-	    break;
+	    goto end;
 	}
 	ret = CardServices(RequestIRQ, link->handle, &link->irq);
 	if (ret != CS_SUCCESS) {
 	    cs_error(RequestIRQ, ret);
-	    break;
+	    goto end;
 	}
 	ret = CardServices(RequestConfiguration, link->handle, &link->conf);
 	if (ret != CS_SUCCESS) {
 	    cs_error(RequestConfiguration, ret);
-	    break;
+	    goto end;
 	}
 	dev->irq = link->irq.AssignedIRQ;
 	irq2dev_map[dev->irq] = dev;
@@ -425,16 +449,10 @@ static void fmvj18x_config(dev_link_t *link)
 	ret = register_netdev(dev);
 	if (ret != 0) {
 	    printk(KERN_NOTICE "fmvj18x_cs: register_netdev() failed\n");
-	    break;
+	    goto end;
 	}
     } while (0);
 
-    /* If any step failed, release any partially configured state */
-    if (ret != 0) {
-	fmvj18x_release((u_long)link);
-	return;
-    }
-
     ioaddr = dev->base_addr;
 
     /* Power On chip and select bank 0 */
@@ -445,20 +463,34 @@ static void fmvj18x_config(dev_link_t *link)
     /* Set hardware address */
     switch (cardtype) {
     case MBH10304:
+    case TDK:
 	do {
 	    u_char *node_id;
 	    
-	    tuple.TupleData = buf;
-	    tuple.TupleDataMax = 64;
+	    tuple.DesiredTuple = CISTPL_FUNCE;
+	    tuple.TupleOffset = 0;
+	    ret = CardServices(GetFirstTuple, handle, &tuple);
+	    if (ret != CS_SUCCESS) goto end;
 	    tuple.TupleOffset = 0;
 	    ret = CardServices(GetTupleData, handle, &tuple);
-	    if (ret != CS_SUCCESS) break;
-	    /* MBH10304's CIS_FUNCE is corrupted */
-	    node_id = &(tuple.TupleData[5]);
+	    if (ret != CS_SUCCESS) goto end;
+	    if (cardtype == MBH10304) {
+		/* MBH10304's CIS_FUNCE is corrupted */
+		node_id = &(tuple.TupleData[5]);
+		card_name = "FMV-J182";
+	    }else {
+		while (tuple.TupleData[0] != CISTPL_FUNCE_LAN_NODE_ID ) {
+		    ret = CardServices(GetNextTuple, handle, &tuple) ;
+		    if (ret != CS_SUCCESS) goto end;
+		    ret = CardServices(GetTupleData, handle, &tuple) ;
+		    if (ret != CS_SUCCESS) goto end;
+		}
+		node_id = &(tuple.TupleData[2]);
+		card_name = "TDK LAK-CD021";
+	    }
 	    /* Read MACID from CIS */
 	    for (i = 0; i < 6; i++)
 		dev->dev_addr[i] = node_id[i];
-	    card_name = "FMV-J182";
 	} while (0);
 	break;
     case MBH10302:
@@ -473,12 +505,20 @@ static void fmvj18x_config(dev_link_t *link)
     link->dev = &((local_info_t *)dev->priv)->node;
     link->state &= ~DEV_CONFIG_PENDING;
 
+    ((struct local_info_t *)dev->priv)->cardtype = cardtype ;
     /* print current configuration */
     printk(KERN_INFO "%s: %s, port %#3lx, irq %d, hw_addr ", 
 	   dev->name, card_name, dev->base_addr, dev->irq);
     for (i = 0; i < 6; i++)
         printk(" %02X", dev->dev_addr[i]);
     printk("\n");
+
+end:
+    /* If any step failed, release any partially configured state */
+    if (ret != 0) {
+	fmvj18x_release((u_long)link);
+	return;
+    }
 } /* fmvj18x_config */
  
 /*====================================================================*/
@@ -654,6 +694,9 @@ static void fjn_interrupt IRQ(int irq, void *dev_id, struct pt_regs *regs)
     /* avoid multiple interrupts */
     outw(0x0000, ioaddr + TX_INTR);
 
+    /* wait for a while */
+    udelay(1);
+
     /* get status */
     tx_stat = inb(ioaddr + TX_STATUS);
     rx_stat = inb(ioaddr + RX_STATUS);
@@ -674,9 +717,11 @@ static void fjn_interrupt IRQ(int irq, void *dev_id, struct pt_regs *regs)
 	fjn_rx(dev);
     }
     if (tx_stat & F_TMT_RDY) {
-	lp->stats.tx_packets++;
+	lp->stats.tx_packets += lp->sent ;
+        lp->sent = 0 ;
 	if (lp->tx_queue) {
 	    outb(DO_TX | lp->tx_queue, ioaddr + TX_START);
+	    lp->sent = lp->tx_queue ;
 	    lp->tx_queue = 0;
 	    lp->tx_queue_len = 0;
 	    dev->trans_start = jiffies;
@@ -734,6 +779,7 @@ static int fjn_start_xmit(struct sk_buff *skb, struct device *dev)
 	lp->tx_started = 0;
 	lp->tx_queue = 0;
 	lp->tx_queue_len = 0;
+	lp->sent = 0;
 	lp->open_time = jiffies;
 	dev->interrupt = 0;
 	dev->tbusy = 0;
@@ -771,7 +817,10 @@ static int fjn_start_xmit(struct sk_buff *skb, struct device *dev)
 
 	/* Disable both interrupts. */
 	outw(0x0000, ioaddr + TX_INTR);
-		
+
+	/* wait for a while */
+	udelay(1);
+
 	outw(length, ioaddr + DATAPORT);
 	outsw(ioaddr + DATAPORT, buf, (length + 1) >> 1);
 
@@ -781,12 +830,13 @@ static int fjn_start_xmit(struct sk_buff *skb, struct device *dev)
 	if (lp->tx_started == 0) {
 	    /* If the Tx is idle, always trigger a transmit. */
 	    outb(DO_TX | lp->tx_queue, ioaddr + TX_START);
+	    lp->sent = lp->tx_queue ;
 	    lp->tx_queue = 0;
 	    lp->tx_queue_len = 0;
 	    dev->trans_start = jiffies;
 	    lp->tx_started = 1;
 	    dev->tbusy = 0;
-	} else if (lp->tx_queue_len < 4096 - 1502)
+	} else if (lp->tx_queue_len < (4096 - (ETH_FRAME_LEN +2)) )
 	    /* Yes, there is room for one more packet. */
 	    dev->tbusy = 0;
 
@@ -803,6 +853,7 @@ static int fjn_start_xmit(struct sk_buff *skb, struct device *dev)
 
 static void fjn_reset(struct device *dev)
 {
+    struct local_info_t *lp = (struct local_info_t *)dev->priv;
     int ioaddr = dev->base_addr;
     int i;
 
@@ -841,7 +892,10 @@ static void fjn_reset(struct device *dev)
     outb(BANK_2, ioaddr + CONFIG_1);
 
     /* set 16col ctrl bits */
-    outb(AUTO_MODE, ioaddr + COL_CTRL);
+    if( lp->cardtype == TDK ) 
+        outb(TDK_AUTO_MODE, ioaddr + COL_CTRL);
+    else
+        outb(AUTO_MODE, ioaddr + COL_CTRL);
 
     /* clear Reserved Regs */
     outb(0x00, ioaddr + BMPR12);
@@ -947,7 +1001,7 @@ static void fjn_rx(struct device *dev)
     /* If any worth-while packets have been received, dev_rint()
 	   has done a mark_bh(NET_BH) for us and will work on them
 	   when we get to the bottom-half routine. */
-    {
+    if( lp->cardtype != TDK ) {
 	int i;
 	for (i = 0; i < 20; i++) {
 	    if ((inb(ioaddr + RX_MODE) & F_BUF_EMP) == F_BUF_EMP)
