@@ -9,13 +9,13 @@
     
 ======================================================================*/
 
-#include <linux/config.h>
+#include "config.h"
+#include "k_compat.h"
 
 /* #define PSYCHO_DEBUG */
 
 #ifdef MODULE
-#include <linux/module.h>
-#include <linux/version.h>
+#define init_ftl_cs init_module
 #endif
 
 #include <linux/kernel.h>
@@ -52,14 +52,14 @@ static int major_dev = 0;
 
 #define DEVICE_NR(minor)	((minor)>>3)
 #define REGION_NR(minor)	((minor)&7)
-#define MINOR_NR(dev,attr,rgn)	(((dev)<<3)+((attr)<<2)+(rgn))
+#define MINOR_NR(dev,rgn)	(((dev)<<3)+(rgn))
 
-#include <drivers/block/blk.h>
+#include BLK_DEV_HDR
 
 #ifdef PCMCIA_DEBUG
 static int pc_debug = PCMCIA_DEBUG;
 static char *version =
-"ftl_cs.c 1.5 1995/07/31 04:58:43 (David Hinds)\n";
+"ftl_cs.c 1.18 1996/05/28 15:16:02 (David Hinds)";
 #endif
 
 static int shuffle_freq = 50;
@@ -88,6 +88,7 @@ static void do_ftl_request(void);
 
 /* Each memory region corresponds to a minor device */
 typedef struct partition_t {
+    dev_node_t		dev;
     u_long		state;
     u_long		*VirtualBlockMap;
     u_long		*VirtualPageMap;
@@ -133,7 +134,7 @@ typedef struct ftl_dev_t {
 static dev_info_t dev_info = "ftl_cs";
 static dev_link_t *dev_table[MAX_DEV] = { NULL, /* ... */ };
 
-static int ftl_blocksizes[MINOR_NR(MAX_DEV, 0, 0)] =
+static int ftl_blocksizes[MINOR_NR(MAX_DEV, 0)] =
 { 0, /* ... */ };
 
 /*====================================================================*/
@@ -158,53 +159,11 @@ static struct file_operations ftl_blk_fops = {
 
 /*====================================================================*/
 
-static void kmsg(char *fmt, ...)
-{
-    char buf[256];
-    va_list args;
-    
-    va_start(args, fmt);
-    vsprintf(buf, fmt, args);
-    va_end(args);
-    printk("%s: %s\n", (char *)dev_info, buf);
-}
-
 static void cs_error(int func, int ret)
 {
     CardServices(ReportError, dev_info, (void *)func, (void *)ret);
 }
 
-/*======================================================================
-
-    If we're linked into the kernel, this gets called from init.c.
-    
-======================================================================*/
-
-#ifndef MODULE
-u_long ftl_init(u_long kmem_start, u_long kmem_end)
-{
-    int i;
-    
-#ifdef PCMCIA_DEBUG
-    if (pc_debug)
-	printk(version);
-#endif
-    
-    register_pcmcia_driver(&dev_info, &ftl_attach, &ftl_detach);
-    
-    major_dev = register_blkdev(0, "ftl", &ftl_blk_fops);
-    if (major_dev == 0)
-	kmsg("unable to grab major device number!");
-    
-    blk_dev[major_dev].request_fn = DEVICE_REQUEST;
-    for (i = 0; i < MINOR_NR(MAX_DEV, 0, 0); i++)
-	ftl_blocksizes[i] = 1024;
-    blksize_size[major_dev] = ftl_blocksizes;
-    
-    return kmem_start;
-}
-#endif
-
 /*======================================================================
 
     ftl_attach() creates an "instance" of the driver, allocating
@@ -223,13 +182,13 @@ static dev_link_t *ftl_attach(void)
     
 #ifdef PCMCIA_DEBUG
     if (pc_debug)
-	kmsg("ftl_attach()");
+	printk(KERN_DEBUG "ftl_cs: ftl_attach()\n");
 #endif
 
     for (i = 0; i < MAX_DEV; i++)
 	if (dev_table[i] == NULL) break;
     if (i == MAX_DEV) {
-	kmsg("no devices available");
+	printk(KERN_NOTICE "ftl_cs: no devices available\n");
 	return NULL;
     }
     
@@ -293,7 +252,7 @@ static void ftl_detach(dev_link_t *link)
 
 #ifdef PCMCIA_DEBUG
     if (pc_debug)
-	kmsg("ftl_detach(0x%p)", link);
+	printk(KERN_DEBUG "ftl_cs: ftl_detach(0x%p)\n", link);
 #endif
     
     /* Locate device structure */
@@ -336,53 +295,48 @@ static void ftl_config(dev_link_t *link)
     ftl_dev_t *dev;
     partition_t *minor;
     region_info_t region;
-    int i, attr, ret, nr[2];
+    dev_node_t **tail;
+    int i, ret, nr;
 
 #ifdef PCMCIA_DEBUG
     if (pc_debug)
-	kmsg("ftl_config(0x%p)", link);
+	printk(KERN_DEBUG "ftl_cs: ftl_config(0x%p)\n", link);
 #endif
 
     /* Configure card */
     link->state |= DEV_CONFIG;
 
-    dev = (ftl_dev_t *)link->priv;
-    for (attr = 0; attr < 2; attr++) {
-	minor = dev->minor + attr*CISTPL_MAX_DEVICES;
-	nr[attr] = 0;
-	region.Attributes =
-	    (attr) ? REGION_TYPE_AM : REGION_TYPE_CM;
-	ret = CardServices(GetFirstRegion, link->handle, &region);
-	while (ret == CS_SUCCESS) {
-	    minor->region = region;
-	    minor++; nr[attr]++;
-	    ret = CardServices(GetNextRegion, link->handle, &region);
-	}
-    }
-    
     for (i = 0; i < MAX_DEV; i++)
 	if (dev_table[i] == link) break;
-    sprintf(link->dev_name, "ftl%d", i);
-    link->minor = MINOR_NR(i, 0, 0);
-    link->major = major_dev;
+    dev = (ftl_dev_t *)link->priv;
+    tail = &link->dev;
+    minor = dev->minor;
+    nr = 0;
+    region.Attributes = REGION_TYPE_CM;
+    ret = CardServices(GetFirstRegion, link->handle, &region);
+    while (ret == CS_SUCCESS) {
+	minor->region = region;
+	sprintf(minor->dev.dev_name, "ftl%dc%d", i, nr);
+	minor->dev.major = major_dev;
+	minor->dev.minor = MINOR_NR(i, nr);
+	*tail = &minor->dev; tail = &minor->dev.next;
+	minor++; nr++;
+	ret = CardServices(GetNextRegion, link->handle, &region);
+    }
+    *tail = NULL;
+    
     link->state &= ~DEV_CONFIG_PENDING;
     
-    if ((nr[0] == 0) && (nr[1] == 0))
-	kmsg("no regions found!");
+    if (nr == 0)
+	printk(KERN_NOTICE "ftl_cs: no regions found!\n");
     else {
-	printk("ftl_cs: %s:", link->dev_name);
-	for (attr = 0; attr < 2; attr++) {
-	    minor = dev->minor + attr*CISTPL_MAX_DEVICES;
-	    if (attr && nr[0] && nr[1])
-		printk(",");
-	    if (nr[attr])
-		printk(" %s", attr ? "attribute" : "common");
-	    for (i = 0; i < nr[attr]; i++) {
-		if (minor[i].region.RegionSize & 0xfffff)
-		    printk(" %ld kb", minor[i].region.RegionSize >> 10);
-		else
-		    printk(" %ld mb", minor[i].region.RegionSize >> 20);
-	    }
+	printk(KERN_INFO "ftl_cs: ftl%d:", i);
+	minor = dev->minor;
+	for (i = 0; i < nr; i++) {
+	    if (minor[i].region.RegionSize & 0xfffff)
+		printk(" %ld kb", minor[i].region.RegionSize >> 10);
+	    else
+		printk(" %ld mb", minor[i].region.RegionSize >> 20);
 	}
 	printk("\n");
     }
@@ -400,18 +354,26 @@ static void ftl_config(dev_link_t *link)
 static void ftl_release(u_long arg)
 {
     dev_link_t *link = (dev_link_t *)arg;
-
+    int i;
+    
 #ifdef PCMCIA_DEBUG
     if (pc_debug)
-	kmsg("ftl_release(0x%p)", link);
+	printk(KERN_DEBUG "ftl_cs: ftl_release(0x%p)\n", link);
 #endif
 
+    for (i = 0; i < MAX_DEV; i++)
+	if (dev_table[i] == link) break;
     if (link->open) {
-	kmsg("release postponed, '%s' still open", link->dev_name);
+#ifdef PCMCIA_DEBUG
+	if (pc_debug)
+	    printk(KERN_DEBUG "ftl_cs: release postponed, 'ftl%d'"
+		   " still open\n", i);
+#endif
 	link->state |= DEV_STALE_CONFIG;
 	return;
     }
 
+    link->dev = NULL;
     if (link->win)
 	CardServices(ReleaseWindow, link->win);
     link->state &= ~DEV_CONFIG;
@@ -438,20 +400,20 @@ static int ftl_event(event_t event, int priority,
 
 #ifdef PCMCIA_DEBUG
     if (pc_debug)
-	kmsg("ftl_event()");
+	printk(KERN_DEBUG "ftl_cs: ftl_event()\n");
 #endif
     
     switch (event) {
 #ifdef PCMCIA_DEBUG
     case CS_EVENT_REGISTRATION_COMPLETE:
 	if (pc_debug)
-	    kmsg("registration complete");
+	    printk(KERN_DEBUG "ftl_cs: registration complete\n");
 	break;
 #endif
     case CS_EVENT_CARD_REMOVAL:
 	link->state &= ~DEV_PRESENT;
 	if (link->state & DEV_CONFIG) {
-	    link->release.expires = 5;
+	    link->release.expires = RUN_AT(HZ/20);
 	    add_timer(&link->release);
 	}
 	break;
@@ -510,13 +472,13 @@ static int scan_partition(partition_t *part)
 	if (strcmp(header.DataOrgTuple+3, "FTL100") == 0) break;
     }
     if (req.Offset == 0x100000) {
-	kmsg("FTL header not found.");
+	printk(KERN_NOTICE "ftl_cs: FTL header not found.\n");
 	return -1;
     }
     if ((header.NumEraseUnits > 65536) || (header.BlockSize != 9) ||
 	(header.EraseUnitSize < 10) || (header.EraseUnitSize > 31) ||
 	(header.NumTransferUnits >= header.NumEraseUnits)) {
-	kmsg("FTL header corrupt!");
+	printk(KERN_NOTICE "ftl_cs: FTL header corrupt!\n");
 	return -1;
     }
     part->header = header;
@@ -551,7 +513,8 @@ static int scan_partition(partition_t *part)
 	}
 	else {
 	    if (xtrans == part->header.NumTransferUnits) {
-		kmsg("format error: too many transfer units!");
+		printk(KERN_NOTICE "ftl_cs: format error: too many "
+		       "transfer units!\n");
 		return -1;
 	    }
 	    if (hdr_ok && (header.LogicalEUN == 0xffff)) {
@@ -572,7 +535,8 @@ static int scan_partition(partition_t *part)
     header = part->header;
     if ((xtrans != header.NumTransferUnits) ||
 	(xvalid+xtrans != header.NumEraseUnits)) {
-	kmsg("format error: erase units don't add up!");
+	printk(KERN_NOTICE "ftl_cs: format error: erase units "
+	       "don't add up!\n");
 	return -1;
     }
     
@@ -632,7 +596,8 @@ static int erase_xfer(ftl_dev_t *dev, partition_t *part,
     xfer = &part->XferInfo[xfernum];
 #ifdef PCMCIA_DEBUG
     if (pc_debug)
-	kmsg("erasing xfer unit at 0x%lx", xfer->Offset);
+	printk(KERN_DEBUG "ftl_cs: erasing xfer unit at 0x%lx\n",
+	       xfer->Offset);
 #endif
     xfer->state = XFER_ERASING;
     /* Is there a free erase slot? */
@@ -642,7 +607,7 @@ static int erase_xfer(ftl_dev_t *dev, partition_t *part,
 	if (i < MAX_ERASE) break;
 #ifdef PCMCIA_DEBUG
 	if (pc_debug)
-	    kmsg("erase queue is full");
+	    printk(KERN_DEBUG "ftl_cs: erase queue is full\n");
 #endif
 	sleep_on(&dev->erase_pending);
     }
@@ -680,7 +645,8 @@ static void save_status(eraseq_entry_t *erase)
     for (i = 0; i < part->header.NumTransferUnits; i++)
 	if (part->XferInfo[i].Offset == erase->Offset) break;
     if (i == part->header.NumTransferUnits) {
-	kmsg("internal error: erase lookup failed!");
+	printk(KERN_NOTICE "ftl_cs: internal error: "
+	       "erase lookup failed!\n");
 	return;
     }
     xfer = &part->XferInfo[i];
@@ -688,7 +654,8 @@ static void save_status(eraseq_entry_t *erase)
 	xfer->state = XFER_ERASED;
     else {
 	xfer->state = XFER_FAILED;
-	kmsg("erase failed: state = %d", erase->State);
+	printk(KERN_NOTICE "ftl_cs: erase failed: state = %d\n",
+	       erase->State);
     }
 }
 
@@ -705,7 +672,8 @@ static void prepare_xfer(partition_t *part, int i)
     
 #ifdef PCMCIA_DEBUG
     if (pc_debug)
-	kmsg("preparing xfer unit at 0x%lx", xfer->Offset);
+	printk(KERN_DEBUG "ftl_cs: preparing xfer unit at 0x%lx\n",
+	       xfer->Offset);
 #endif
 
     /* Write the transfer unit header */
@@ -765,7 +733,8 @@ static int copy_erase_unit(partition_t *part, u_short srcunit,
     xfer = &part->XferInfo[xferunit];
 #ifdef PCMCIA_DEBUG
     if (pc_debug)
-	kmsg("copying block 0x%lx to 0x%lx", eun->Offset, xfer->Offset);
+	printk(KERN_DEBUG "ftl_cs: copying block 0x%lx to 0x%lx\n",
+	       eun->Offset, xfer->Offset);
 #endif
 	
     req.Attributes = MEM_OP_BUFFER_KERNEL;
@@ -878,7 +847,7 @@ static int reclaim_block(ftl_dev_t *dev, partition_t *part)
 
 #ifdef PCMCIA_DEBUG
     if (pc_debug)
-	kmsg("reclaiming space...");
+	printk(KERN_DEBUG "ftl_cs: reclaiming space...\n");
 #endif
 	
     /* Pick the least erased transfer unit */
@@ -902,14 +871,16 @@ static int reclaim_block(ftl_dev_t *dev, partition_t *part)
 	    if (queued) {
 #ifdef PCMCIA_DEBUG
 		if (pc_debug)
-		    kmsg("waiting for transfer unit to be prepared...");
+		    printk(KERN_DEBUG "ftl_cs: waiting for transfer "
+			   "unit to be prepared...\n");
 #endif
 		sleep_on(&dev->erase_pending);
 	    }
 	    else {
 		static int ne = 0;
 		if (++ne < 5)
-		    kmsg("reclaim failed: no suitable transfer units!");
+		    printk(KERN_NOTICE "ftl_cs: reclaim failed: no "
+			   "suitable transfer units!\n");
 		return CS_GENERAL_FAILURE;
 	    }
 	}
@@ -919,7 +890,7 @@ static int reclaim_block(ftl_dev_t *dev, partition_t *part)
     if ((jiffies % shuffle_freq) == 0) {
 #ifdef PCMCIA_DEBUG
 	if (pc_debug)
-	    kmsg("recycling freshest block...");
+	    printk(KERN_DEBUG "ftl_cs: recycling freshest block...\n");
 #endif
 	best = 0xffffffff;
 	for (i = 0; i < part->DataUnits; i++)
@@ -938,7 +909,8 @@ static int reclaim_block(ftl_dev_t *dev, partition_t *part)
 	if (best == 0) {
 	    static int ne = 0;
 	    if (++ne < 5)
-		kmsg("reclaim failed: no free blocks!");
+		printk(KERN_NOTICE "ftl_cs: reclaim failed: "
+		       "no free blocks!\n");
 	    return CS_GENERAL_FAILURE;
 	}
     }
@@ -946,7 +918,7 @@ static int reclaim_block(ftl_dev_t *dev, partition_t *part)
     if (ret == CS_SUCCESS)
 	erase_xfer(dev, part, xfer);
     else
-	kmsg("copy_erase_unit failed!");
+	printk(KERN_NOTICE "ftl_cs: copy_erase_unit failed!\n");
     return ret;
 } /* reclaim_block */
 
@@ -964,11 +936,12 @@ static int reclaim_block(ftl_dev_t *dev, partition_t *part)
 static void dump_lists(partition_t *part)
 {
     int i;
-    kmsg("Free total = %d", part->FreeTotal);
+    printk(KERN_DEBUG "ftl_cs: Free total = %d\n", part->FreeTotal);
     for (i = 0; i < part->DataUnits; i++)
-	kmsg("  unit %d: %d phys, %d free, %d deleted", i,
-	     part->EUNInfo[i].Offset >> part->header.EraseUnitSize,
-	     part->EUNInfo[i].Free, part->EUNInfo[i].Deleted);
+	printk(KERN_DEBUG "ftl_cs:   unit %d: %d phys, %d free, "
+	       "%d deleted\n", i,
+	       part->EUNInfo[i].Offset >> part->header.EraseUnitSize,
+	       part->EUNInfo[i].Free, part->EUNInfo[i].Deleted);
 }
 #endif
 
@@ -1016,12 +989,13 @@ static u_long find_free(partition_t *part)
 	if (++ne == 1)
 	    dump_lists(part);
 #endif
-	kmsg("bad free list!");
+	printk(KERN_NOTICE "ftl_cs: bad free list!\n");
 	return 0;
     }
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 1)
-	kmsg("found free block at %ld in %d", blk, eun);
+	printk(KERN_DEBUG "ftl_cs: found free block at %ld in %d\n",
+	       blk, eun);
 #endif
     return blk;
     
@@ -1045,7 +1019,7 @@ static int ftl_open(struct inode *inode, struct file *file)
     
 #ifdef PCMCIA_DEBUG
     if (pc_debug)
-	kmsg("ftl_open(%d)", minor);
+	printk(KERN_DEBUG "ftl_cs: ftl_open(%d)\n", minor);
 #endif
 
     link = dev_table[DEVICE_NR(minor)];
@@ -1065,21 +1039,19 @@ static int ftl_open(struct inode *inode, struct file *file)
 	    return -ENODEV;
 	if (scan_partition(partition) == 0) {
 	    partition->state = FTL_FORMATTED;
-	    kmsg("opening %d kb FTL partition",
-		 partition->header.FormattedSize >> 10);
+	    printk(KERN_INFO "ftl_cs: opening %ld kb FTL partition\n",
+		   partition->header.FormattedSize >> 10);
 	}
 	else {
 	    CardServices(CloseMemory, partition->handle);
-	    kmsg("FTL partition is invalid.");
+	    printk(KERN_NOTICE "ftl_cs: FTL partition is invalid.\n");
 	    return -ENODEV;
 	}
     }
     
     partition->open++;
     link->open++;
-#ifdef MODULE
     MOD_INC_USE_COUNT;
-#endif
     return 0;
 } /* ftl_open */
 
@@ -1095,7 +1067,7 @@ static void ftl_close(struct inode *inode, struct file *file)
     
 #ifdef PCMCIA_DEBUG
     if (pc_debug)
-	kmsg("ftl_close(%d)", minor);
+	printk(KERN_DEBUG "ftl_cs: ftl_close(%d)\n", minor);
 #endif
 
     /* Flush all writes */
@@ -1140,9 +1112,7 @@ static void ftl_close(struct inode *inode, struct file *file)
 	}
     }
     
-#ifdef MODULE
     MOD_DEC_USE_COUNT;
-#endif
 
 } /* ftl_close */
 
@@ -1161,11 +1131,11 @@ static int ftl_read(partition_t *part, caddr_t buffer,
     
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 1)
-	kmsg("ftl_read(0x%p, 0x%lx, %ld)",
-	     part->handle, sector, nblocks);
+	printk(KERN_DEBUG "ftl_cs: ftl_read(0x%p, 0x%lx, %ld)\n",
+	       part->handle, sector, nblocks);
 #endif
     if (!(part->state & FTL_FORMATTED)) {
-	kmsg("bad partition");
+	printk(KERN_NOTICE "ftl_cs: bad partition\n");
 	return -EIO;
     }
     bsize = part->region.BlockSize;
@@ -1173,7 +1143,7 @@ static int ftl_read(partition_t *part, caddr_t buffer,
     req.Count = SECTOR_SIZE;
     for (i = 0; i < nblocks; i++) {
 	if (((sector+i) * SECTOR_SIZE) >= part->header.FormattedSize) {
-	    kmsg("bad read offset");
+	    printk(KERN_NOTICE "ftl_cs: bad read offset\n");
 	    return -EIO;
 	}
 	log_addr = part->VirtualBlockMap[sector+i];
@@ -1212,8 +1182,8 @@ static int set_bam_entry(partition_t *part, u_long log_addr,
     
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 1)
-	kmsg("set_bam_entry(0x%p, 0x%lx, 0x%lx)",
-	     part->handle, log_addr, virt_addr);
+	printk(KERN_DEBUG "ftl_cs: set_bam_entry(0x%p, 0x%lx, 0x%lx)\n",
+	       part->handle, log_addr, virt_addr);
 #endif
     bsize = part->region.BlockSize;
     eun = log_addr / bsize;
@@ -1230,9 +1200,9 @@ static int set_bam_entry(partition_t *part, u_long log_addr,
 	(!BLOCK_DELETED(virt_addr) && (old_addr != 0xfffffffe))) {
 	static int ne = 0;
 	if (++ne < 5) {
-	    kmsg("set_bam_entry() inconsistency!");
-	    kmsg("  log_addr = 0x%lx, old = 0x%lx, new = 0x%lx",
-		 log_addr, old_addr, virt_addr);
+	    printk(KERN_NOTICE "ftl_cs: set_bam_entry() inconsistency!\n");
+	    printk(KERN_NOTICE "ftl_cs:   log_addr = 0x%lx, old = 0x"
+		   "%lx, new = 0x%lx\n", log_addr, old_addr, virt_addr);
 	}
 	return CS_GENERAL_FAILURE;
     }
@@ -1242,9 +1212,11 @@ static int set_bam_entry(partition_t *part, u_long log_addr,
 	if (part->bam_cache[blk] != old_addr) {
 	    static int ne = 0;
 	    if (++ne < 5) {
-		kmsg("set_bam_entry() inconsistency!");
-		kmsg("  log_addr = 0x%lx, cache = 0x%lx, card = 0x%lx",
-		     part->bam_cache[blk], old_addr);
+		printk(KERN_NOTICE "ftl_cs: set_bam_entry() "
+		       "inconsistency!\n");
+		printk(KERN_NOTICE "ftl_cs:   log_addr = 0x%lx, cache"
+		       " = 0x%lx, card = 0x%lx\n",
+		       part->bam_cache[blk], old_addr);
 	    }
 	    return CS_GENERAL_FAILURE;
 	}
@@ -1255,9 +1227,9 @@ static int set_bam_entry(partition_t *part, u_long log_addr,
     ret = CardServices(WriteMemory, part->handle, &req, &virt_addr);
 #ifdef PSYCHO_DEBUG
     if (ret != CS_SUCCESS) {
-	kmsg("set_bam_entry() failed!");
-	kmsg("  log_addr = 0x%lx, old = 0x%lx, new = 0x%lx",
-	     log_addr, old_addr, virt_addr);
+	printk(KERN_NOTICE "ftl_cs: set_bam_entry() failed!\n");
+	printk(KERN_NOTICE "ftl_cs:   log_addr = 0x%lx, old = 0x%lx,"
+	       " new = 0x%lx\n", log_addr, old_addr, virt_addr);
 	cs_error(WriteMemory, ret);
     }
 #endif
@@ -1273,11 +1245,11 @@ static int ftl_write(ftl_dev_t *dev, partition_t *part, caddr_t buffer,
 
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 1)
-	kmsg("ftl_write(0x%p, %ld, %ld)",
-	     part->handle, sector, nblocks);
+	printk(KERN_DEBUG "ftl_cs: ftl_write(0x%p, %ld, %ld)\n",
+	       part->handle, sector, nblocks);
 #endif
     if (!(part->state & FTL_FORMATTED)) {
-	kmsg("bad partition");
+	printk(KERN_NOTICE "ftl_cs: bad partition\n");
 	return -EIO;
     }
     /* See if we need to reclaim space, before we start */
@@ -1293,7 +1265,7 @@ static int ftl_write(ftl_dev_t *dev, partition_t *part, caddr_t buffer,
     virt_addr = sector * SECTOR_SIZE | BLOCK_DATA;
     for (i = 0; i < nblocks; i++) {
 	if (virt_addr >= part->header.FormattedSize) {
-	    kmsg("bad write offset");
+	    printk(KERN_NOTICE "ftl_cs: bad write offset\n");
 	    return -EIO;
 	}
 
@@ -1302,7 +1274,8 @@ static int ftl_write(ftl_dev_t *dev, partition_t *part, caddr_t buffer,
 	if (blk == 0) {
 	    static int ne = 0;
 	    if (++ne < 5)
-		kmsg("internal error: no free blocks!");
+		printk(KERN_NOTICE "ftl_cs: internal error: "
+		       "no free blocks!\n");
 	    return -ENOSPC;
 	}
 
@@ -1317,9 +1290,10 @@ static int ftl_write(ftl_dev_t *dev, partition_t *part, caddr_t buffer,
 		      blk * SECTOR_SIZE);
 	ret = CardServices(WriteMemory, part->handle, &req, buffer);
 	if (ret != CS_SUCCESS) {
-	    kmsg("block write failed!");
-	    kmsg("  log_addr = 0x%lx, virt_addr = 0x%lx, Offset = 0x%lx",
-		 log_addr, virt_addr, req.Offset);
+	    printk(KERN_NOTICE "ftl_cs: block write failed!\n");
+	    printk(KERN_NOTICE "ftl_cs:   log_addr = 0x%lx, virt_addr"
+		   " = 0x%lx, Offset = 0x%lx\n", log_addr, virt_addr,
+		   req.Offset);
 	    cs_error(WriteMemory, ret);
 	    return -EIO;
 	}
@@ -1398,14 +1372,15 @@ static void do_ftl_request(void)
 
 #ifdef PCMCIA_DEBUG
     if (pc_debug)
-	kmsg("starting do_ftl_request()");
+	printk(KERN_DEBUG "ftl_cs: starting do_ftl_request()\n");
 #endif
     
     do {
 	sti();
 	INIT_REQUEST;
 
-	minor = MINOR(CURRENT->dev);
+	minor = MINOR(DEVICE(CURRENT));
+	
 	link = dev_table[DEVICE_NR(minor)];
 	dev = (ftl_dev_t *)link->priv;
 	part = &dev->minor[REGION_NR(minor)];
@@ -1441,22 +1416,20 @@ static void do_ftl_request(void)
 
 /*====================================================================*/
 
-#ifdef MODULE
-char kernel_version[] = UTS_RELEASE;
-
-int init_module(void)
+int init_ftl_cs(void)
 {
     servinfo_t serv;
     int i;
     
 #ifdef PCMCIA_DEBUG
     if (pc_debug)
-	printk(version);
+	printk(KERN_INFO "%s\n", version);
 #endif
     
     CardServices(GetCardServicesInfo, &serv);
     if (serv.Revision != CS_RELEASE_CODE) {
-	kmsg("Card Services release does not match!");
+	printk(KERN_NOTICE "ftl_cs: Card Services release "
+	       "does not match!\n");
 	return -1;
     }
     
@@ -1464,22 +1437,27 @@ int init_module(void)
 
     major_dev = register_blkdev(0, "ftl", &ftl_blk_fops);
     if (major_dev == 0)
-	kmsg("unable to grab major device number!");
+	printk(KERN_NOTICE "ftl_cs: unable to grab major "
+	       "device number!\n");
 
     blk_dev[major_dev].request_fn = DEVICE_REQUEST;
-    for (i = 0; i < MINOR_NR(MAX_DEV, 0, 0); i++)
+    for (i = 0; i < MINOR_NR(MAX_DEV, 0); i++)
 	ftl_blocksizes[i] = 1024;
     blksize_size[major_dev] = ftl_blocksizes;
     
     return 0;
 }
 
+#ifdef MODULE
 void cleanup_module(void)
 {
     int i;
     dev_link_t *link;
-    
-    kmsg("unloading");
+
+#ifdef PCMCIA_DEBUG
+    if (pc_debug)
+	printk(KERN_DEBUG "ftl_cs: unloading\n");
+#endif
     unregister_pcmcia_driver(&dev_info);
     if (major_dev != 0)
 	unregister_blkdev(major_dev, "ftl");
