@@ -43,7 +43,7 @@
 #ifdef PCMCIA_DEBUG
 static int pc_debug = PCMCIA_DEBUG;
 static const char *version =
-"i82365.c 1.95 1996/05/31 05:28:24 (David Hinds)";
+"i82365.c 1.99 1996/08/15 03:16:51 (David Hinds)";
 #endif
 
 /*====================================================================*/
@@ -59,8 +59,11 @@ static int ignore = -1;
 /* Should we probe at 0x3e2 for an extra controller? */
 static int extra_sockets = 0;
 
-/* Bit map of interrupts to choose from -- 0 means just autoprobe */
-static u_long irq_mask = 0;
+/* Probe for safe interrupts? */
+static int do_scan = 1;
+
+/* Bit map of interrupts to choose from */
+static u_long irq_mask = 0xffff;
 
 /* The card status change interrupt -- 0 means autoselect */
 static int cs_irq = 0;
@@ -283,7 +286,7 @@ static int identify(int sock)
 	type = IS_I82365B; break;
     case 0x84:
 	type = IS_VLSI; break;
-    case 0x88: case 0x89:
+    case 0x88: case 0x89: case 0x8a:
 	type = IS_IBM; break;
     case 0x8b: case 0x8c:
 	/* Special: fall into Vadem test */
@@ -299,7 +302,7 @@ static int identify(int sock)
 	i365_bset(sock, VG468_MISC, VG468_MISC_VADEMREV);
 	val = i365_get(sock, I365_IDENT);
 	if (val & I365_IDENT_VADEM) {
-	    type = ((val & 7) == 4) ? IS_VG469 : IS_VG468;
+	    type = ((val & 7) >= 4) ? IS_VG469 : IS_VG468;
 	    i365_bclr(sock, VG468_MISC, VG468_MISC_VADEMREV);
 	}
 	else if (type == IS_VG468)
@@ -324,7 +327,7 @@ static int identify(int sock)
 
 /*====================================================================*/
 
-void add_pcic(int type, int ns)
+static void add_pcic(int type, int ns)
 {
     int i, j, irq;
     u_short base;
@@ -365,8 +368,7 @@ void add_pcic(int type, int ns)
 
     /* Build interrupt mask; scan if possible */
     printk(KERN_INFO "  irq mask (");
-    mask = i365_cap.irq_mask;
-    if (irq_mask) mask &= irq_mask;
+    mask = i365_cap.irq_mask & irq_mask;
     switch (type) {
     case IS_PD6710:
     case IS_PD6729:
@@ -377,22 +379,21 @@ void add_pcic(int type, int ns)
 	printk("Cirrus");
 	break;
     default:
-	scan = irq_scan(base, mask);
-	if (scan) {
-	    u_long tmp;
+	if (do_scan && (scan = irq_scan(base, mask))) {
 	    mask = scan;
 	    printk("scanned");
-	    /* Check for only two interrupts available */
-	    tmp = (scan & (scan-1));
-	    if (((tmp & (tmp-1)) == 0) && (poll_interval == 0))
-		poll_interval = HZ;
 	}
-	else
+	else 
 	    printk("default");
 	break;
     }
     printk(") = 0x%lx, ", mask);
     
+    /* Check for only two interrupts available */
+    scan = (mask & (mask-1));
+    if (((scan & (scan-1)) == 0) && (poll_interval == 0))
+	poll_interval = HZ;
+
     /* Pick a card status change interrupt */
     if (poll_interval == 0) {
 	if (cs_irq == 0) {
@@ -403,7 +404,7 @@ void add_pcic(int type, int ns)
 	    irq = i;
 	}
 	else {
-	    if (REQUEST_IRQ(cs_irq, i365_interrupt, 0, "i82365") != 0)
+	    if (REQUEST_IRQ(cs_irq, i365_interrupt, 0, "i82365") == 0)
 		irq = cs_irq;
 	    else
 		irq = 0;
@@ -422,9 +423,9 @@ void add_pcic(int type, int ns)
     for (i = base; i < base+ns; i++) {
 	socket_table[i].cap.irq_mask = mask;
 	socket_table[i].cs_irq = irq;
-	if (cs_irq) {
+	if (irq) {
 	    j = socket_table[i].psock;
-	    i365_set(j, I365_CSCINT, (cs_irq << 4));
+	    i365_set(j, I365_CSCINT, (irq << 4));
 	    i365_bclr(j, I365_INTCTL, I365_INTR_ENA);
 	    i365_get(j, I365_CSC);
 	}
