@@ -1,6 +1,6 @@
 /*======================================================================
 
-    A simple MTD for Intel Series 2+ Flash devices
+    A simple MTD for Intel Series 2 Flash devices
 
     Written by David Hinds, dhinds@allegro.stanford.edu
 
@@ -11,13 +11,13 @@
     
 ======================================================================*/
 
-#include <linux/config.h>
+#include "config.h"
+#include "k_compat.h"
 
 /* #define PCMCIA_DEBUG 1 */
 
 #ifdef MODULE
-#include <linux/module.h>
-#include <linux/version.h>
+#define init_iflash2_mtd init_module
 #endif
 
 #include <linux/kernel.h>
@@ -41,18 +41,21 @@
 #include "cisreg.h"
 #include "cistpl.h"
 #include "ds.h"
+#include "mem_op.h"
 #include "iflash.h"
 
 #ifdef PCMCIA_DEBUG
 static int pc_debug = PCMCIA_DEBUG;
 static char *version =
-"iflash2_mtd.c 1.9 1995/07/31 04:04:11 (David Hinds)\n";
+"iflash2_mtd.c 1.20 1996/05/28 15:14:24 (David Hinds)";
 #endif
 
-static int vpp_timeout_period	= 100;	/* in ticks */
-static int vpp_settle		= 10;	/* in ticks */
-static int erase_timeout	= 100;	/* in ms */
-static int erase_limit		= 1000;	/* in ticks */
+/*====================================================================*/
+
+static int vpp_timeout_period	= HZ;		/* in ticks */
+static int vpp_settle		= HZ/10;	/* in ticks */
+static int erase_timeout	= 100;		/* in ms */
+static int erase_limit		= 10*HZ;	/* in ticks */
 
 static int retry_limit		= 4;	/* max retries */
 
@@ -104,17 +107,6 @@ static dev_link_t *dev_list = NULL;
 
 /*====================================================================*/
 
-static void kmsg(char *fmt, ...)
-{
-    char buf[256];
-    va_list args;
-    
-    va_start(args, fmt);
-    vsprintf(buf, fmt, args);
-    va_end(args);
-    printk("%s: %s\n", (char *)dev_info, buf);
-}
-
 static void cs_error(int func, int ret)
 {
     CardServices(ReportError, dev_info, (void *)func, (void *)ret);
@@ -129,14 +121,18 @@ static void cs_error(int func, int ret)
 static void abort_cmd(dev_link_t *link, volatile caddr_t base,
 		      int cell, mtd_mod_win_t *old)
 {
+    u_char *addr = base + CISREG_SLEEP + ((cell>>3)<<1);
     /* Map the CIS register page into memory */
     mtd_mod_win_t mod = { WIN_MEMORY_TYPE_AM, 250, 0x4000 };
-    kmsg("abort");
+#ifdef PCMCIA_DEBUG
+    if (pc_debug)
+	printk(KERN_DEBUG "iflash2_mtd: abort\n");
+#endif
     MTDHelperEntry(MTDModifyWindow, link->win, &mod);
     /* Put the cell to sleep, then wake it up again */
-    base[CISREG_SLEEP + ((cell>>3)<<1)] |= 1 << (cell & 7);
+    writeb(readb(addr) | (1 << (cell & 7)), addr);
     udelay(10L);
-    base[CISREG_SLEEP + ((cell>>3)<<1)] &= ~(1 << (cell & 7));
+    writeb(readb(addr) & ~(1 << (cell & 7)), addr);
     udelay(10L);
     /* Restore the original memory window mapping */
     MTDHelperEntry(MTDModifyWindow, link->win, old);
@@ -147,14 +143,15 @@ static inline int byte_write(volatile u_char *address, u_char data)
     register u_char CSR = 0;
     register u_short i;
     for (i = 0; i < max_tries; i++) {
-	CSR = *address;
+	CSR = readb(address);
 	if (CSR & LOW(CSR_WR_READY)) {
-	    *address = LOW(IF_WRITE);
-	    *address = data;
+	    writeb(LOW(IF_WRITE), address);
+	    writeb(data, address);
 	    return CS_SUCCESS;
 	}
     }
-    kmsg("byte_write timed out at 0x%p, CSR = 0x%x", address, CSR);
+    printk(KERN_NOTICE "iflash2_mtd: byte_write timed out at 0x%p, "
+	   "CSR = 0x%x\n", address, CSR);
     return CS_WRITE_FAILURE;
 }
 
@@ -174,14 +171,15 @@ __asm__ __volatile__(
     register u_short CSR = 0, i;
 
     for (i = 0; i < max_tries; i++) {
-	CSR = *address;
+	CSR = readw(address);
 	if ((CSR & CSR_WR_READY) == CSR_WR_READY) {
-	    *address = IF_WRITE;
-	    *address = data;
+	    writew(IF_WRITE, address);
+	    writew(data, address);
 	    return CS_SUCCESS;
 	}
     }
-    kmsg("word_write timed out at 0x%p, CSR = 0x%x", address, CSR);
+    printk(KERN_NOTICE "iflash2_mtd: word_write timed out at 0x%p, "
+	   "CSR = 0x%x\n", address, CSR);
     return CS_WRITE_FAILURE;
 #endif
 }
@@ -189,17 +187,19 @@ __asm__ __volatile__(
 static int check_write(volatile u_short *address)
 {
     u_short CSR = 0, i;
-    *address = IF_READ_CSR;
+    writew(IF_READ_CSR, address);
     for (i = 0; i < max_tries; i++) {
-	CSR = *address;
+	CSR = readw(address);
 	if ((CSR & CSR_WR_READY) == CSR_WR_READY) break;
     }
     if (i == max_tries) {
-	kmsg("check_write: timed out!  CSR = 0x%x", CSR);
+	printk(KERN_NOTICE "iflash2_mtd: check_write timed out!"
+	       "  CSR = 0x%x\n", CSR);
 	return CS_GENERAL_FAILURE;
     }
     if (CSR & (CSR_WR_ERR | CSR_VPP_LOW)) {
-	kmsg("write error: CSR = 0x%x", CSR);
+	printk(KERN_NOTICE "iflash2_mtd: write error: CSR = 0x%x\n",
+	       CSR);
 	return CS_WRITE_FAILURE;
     }
     else
@@ -208,19 +208,20 @@ static int check_write(volatile u_short *address)
 
 static void block_erase(volatile u_short *address)
 {
-    *address = IF_BLOCK_ERASE;
-    *address = IF_CONFIRM;
+    writew(IF_BLOCK_ERASE, address);
+    writew(IF_CONFIRM, address);
 }
 
 static int check_erase(volatile u_short *address)
 {
     u_short CSR;
-    *address = IF_READ_CSR;
-    CSR = *address;
+    writew(IF_READ_CSR, address);
+    CSR = readw(address);
     if ((CSR & CSR_WR_READY) != CSR_WR_READY)
 	return CS_BUSY;
     else if (CSR & (CSR_ERA_ERR | CSR_VPP_LOW | CSR_WR_ERR)) {
-	kmsg("erase failed: CSR = 0x%x", CSR);
+	printk(KERN_NOTICE "iflash2_mtd: erase failed: CSR = 0x%x\n",
+	       CSR);
 	return CS_WRITE_FAILURE;
     }
     else
@@ -232,43 +233,46 @@ static int suspend_erase(volatile u_short *address)
     u_short CSR = 0;
     u_long i;
 
-    *address = IF_ERASE_SUSPEND;
-    *address = IF_READ_CSR;
+    writew(IF_ERASE_SUSPEND, address);
+    writew(IF_READ_CSR, address);
     for (i = 0; i < max_tries; i++) {
-	CSR = *address;
+	CSR = readw(address);
 	if ((CSR & CSR_WR_READY) == CSR_WR_READY) break;
     }
     if (i == max_tries) {
-	kmsg("suspend_erase timed out: CSR = 0x%x", CSR);
+	printk(KERN_NOTICE "iflash2_mtd: suspend_erase timed out: "
+	       "CSR = 0x%x\n", CSR);
 	return CS_GENERAL_FAILURE;
     }
-    *address = IF_READ_ARRAY;
+    writew(IF_READ_ARRAY, address);
     return CS_SUCCESS;
 }
 
 static void resume_erase(volatile u_short *address)
 {
     u_short CSR;
-    *address = IF_READ_CSR;
-    CSR = *address;
+    writew(IF_READ_CSR, address);
+    CSR = readw(address);
     /* Only give resume signal if the erase is really suspended */
     if (CSR & CSR_ERA_SUSPEND)
-	*address = IF_CONFIRM;
+	writew(IF_CONFIRM, address);
 }
 
 static void reset_block(volatile u_short *address)
 {
     u_short CSR, i;
-    *address = IF_CLEAR_CSR;
+    writew(IF_CLEAR_CSR, address);
     for (i = 0; i < 100; i++) {
-	*address = IF_READ_CSR;
-	CSR = *address;
+	writew(IF_READ_CSR, address);
+	CSR = readw(address);
 	if (CSR != 0xffff) break;
 	udelay(1000);
     }
+#ifdef PCMCIA_DEBUG
     if (i)
-	kmsg("reset after %d iterations", i);
-    *address = IF_READ_ARRAY;
+	printk(KERN_DEBUG "iflash2_mtd: reset after %d iterations\n", i);
+#endif
+    writew(IF_READ_ARRAY, address);
 }
 
 /*======================================================================
@@ -303,7 +307,7 @@ static int vpp_setup(dev_link_t *link, mtd_request_t *req)
 	    else {
 #ifdef PCMCIA_DEBUG
 		if (pc_debug > 1)
-		    kmsg("raising Vpp...");
+		    printk(KERN_DEBUG "iflash2_mtd: raising Vpp...\n");
 #endif
 		dev->vpp_start = jiffies;
 		vpp_req.Vpp1 = vpp_req.Vpp2 = 120;
@@ -328,7 +332,7 @@ static void vpp_off(u_long arg)
 
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 1)
-	kmsg("lowering Vpp...");
+	printk(KERN_DEBUG "iflash2_mtd: lowering Vpp...\n");
 #endif
     dev = (flash_dev_t *)link->priv;
     dev->vpp_timeout.expires = 0;
@@ -342,29 +346,11 @@ static void vpp_shutdown(dev_link_t *link)
     dev = (flash_dev_t *)link->priv;
     dev->vpp_usage--;
     if (dev->vpp_usage == 0) {
-	dev->vpp_timeout.expires = vpp_timeout_period;
+	dev->vpp_timeout.expires = RUN_AT(vpp_timeout_period);
 	add_timer(&dev->vpp_timeout);
     }
 }
 
-/*======================================================================
-
-    If we're linked into the kernel, this gets called from init.c.
-    
-======================================================================*/
-
-#ifndef MODULE
-u_long flash_init(u_long kmem_start, u_long kmem_end)
-{
-#ifdef PCMCIA_DEBUG
-    if (pc_debug)
-	printk(version);
-#endif
-    register_pcmcia_driver(&dev_info, &flash_attach, &flash_detach);
-    return kmem_start;
-}
-#endif
-
 /*======================================================================
 
     flash_attach() creates an "instance" of the driver, allocating
@@ -382,7 +368,7 @@ static dev_link_t *flash_attach(void)
     
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 1)
-	kmsg("flash_attach()");
+	printk(KERN_DEBUG "iflash2_mtd: flash_attach()\n");
 #endif
 
     /* Create new memory card device */
@@ -435,7 +421,7 @@ static void flash_detach(dev_link_t *link)
 
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 1)
-	kmsg("flash_detach(0x%p)", link);
+	printk(KERN_DEBUG "iflash2_mtd: flash_detach(0x%p)\n", link);
 #endif
     
     /* Locate device structure */
@@ -477,7 +463,9 @@ static void flash_detach(dev_link_t *link)
 
 static void printk_size(u_long sz)
 {
-    if (sz & 0xfffff)
+    if (sz & 0x3ff)
+	printk("%ld bytes", sz);
+    else if (sz & 0x0fffff)
 	printk("%ld kb", sz >> 10);
     else
 	printk("%ld mb", sz >> 20);
@@ -493,7 +481,7 @@ static void flash_config(dev_link_t *link)
 
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 1)
-	kmsg("flash_config(0x%p)", link);
+	printk(KERN_DEBUG "iflash2_mtd: flash_config(0x%p)\n", link);
 #endif
 
     /* Allocate a 4K memory window */
@@ -529,11 +517,14 @@ static void flash_config(dev_link_t *link)
 		kfree_s(dev->flash[i], sizeof(struct flash_region_t));
 		break;
 	    }
-	    printk("iflash2_mtd: %s at 0x%lx, ",
+	    printk(KERN_INFO "iflash2_mtd: %s at 0x%lx, ",
 		   attr ? "attr" : "common", region.CardOffset);
 	    printk_size(region.RegionSize);
 	    printk(", %ld ns\n", region.AccessSpeed);
 	    memset(dev->flash[i], 0, sizeof(struct flash_region_t));
+	    /* Assume 128K blocks, if no geometry info present */
+	    if (region.BlockSize == 1)
+		region.BlockSize = 0x20000;
 	    dev->flash[i]->region = region;
 	    /* All Series 2 cards have 2MB component pairs */
 	    dev->flash[i]->cell_size = 0x200000;
@@ -560,7 +551,7 @@ static void flash_release(u_long arg)
 
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 1)
-	kmsg("flash_release(0x%p)", link);
+	printk(KERN_DEBUG "iflash2_mtd: flash_release(0x%p)\n", link);
 #endif
 
     link->state &= ~DEV_CONFIG;
@@ -602,9 +593,9 @@ static int flash_read(dev_link_t *link, char *buf, mtd_request_t *req)
     
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 1)
-	kmsg("flash_read(0x%p, 0x%ld, 0x%p, 0x%lx, 0x%lx)",
-	     link, req->MediaID, buf, req->SrcCardOffset,
-	     req->TransferLength);
+	printk(KERN_DEBUG "iflash2_mtd: flash_read(0x%p, 0x%ld, 0x%p,"
+	       " 0x%lx, 0x%lx)\n", link, req->MediaID, buf,
+	       req->SrcCardOffset, req->TransferLength);
 #endif
 
     flash = (flash_region_t *)(req->MediaID);
@@ -649,9 +640,9 @@ static int flash_read(dev_link_t *link, char *buf, mtd_request_t *req)
 	nb = (from+length > WINDOW_SIZE) ? WINDOW_SIZE-from : length;
 
 	if (req->Function & MTD_REQ_KERNEL)
-	    memcpy(buf, &dev->Base[from], nb);
+	    memcpy_frompc(buf, &dev->Base[from], nb);
 	else
-	    memcpy_tofs(buf, &dev->Base[from], nb);
+	    memcpy_frompc_tofs(buf, &dev->Base[from], nb);
 	
 	buf += nb;
 	from = 0;
@@ -661,9 +652,10 @@ static int flash_read(dev_link_t *link, char *buf, mtd_request_t *req)
 #ifdef PCMCIA_DEBUG
     time = jiffies - time;
     if (pc_debug && (time > 1))
-	kmsg("read complete, time = %ld, avg = %ld ns/word, rate = "
-	     "%ld kb/sec", time, time*20000000/req->TransferLength,
-	     req->TransferLength*100/(time*1024));
+	printk(KERN_DEBUG "iflash2_mtd: read complete, time = %ld, "
+	       "avg = %ld ns/word, rate = %ld kb/sec\n", time,
+	       time*20000000/req->TransferLength,
+	       req->TransferLength*100/(time*1024));
 #endif
     
 done:
@@ -737,9 +729,9 @@ static int flash_write(dev_link_t *link, char *buf, mtd_request_t *req)
 
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 1)
-	kmsg("flash_write(0x%p, 0x%lx, 0x%p, 0x%lx, 0x%lx)",
-	     link, req->MediaID, buf, req->DestCardOffset,
-	     req->TransferLength);
+	printk(KERN_DEBUG "iflash2_mtd: flash_write(0x%p, 0x%lx, "
+	       "0x%p, 0x%lx, 0x%lx)\n", link, req->MediaID, buf,
+	       req->DestCardOffset, req->TransferLength);
 #endif
 
     /* Check card write protect status */
@@ -792,7 +784,8 @@ static int flash_write(dev_link_t *link, char *buf, mtd_request_t *req)
 	    abort_cmd(link, dev->Base, cell, &mod);
 	}
 	if (retry == retry_limit) {
-	    kmsg("write failed: too many retries!");
+	    printk(KERN_NOTICE "iflash2_mtd: write failed: "
+		   "too many retries!\n");
 	    goto done;
 	}
 	
@@ -804,9 +797,10 @@ static int flash_write(dev_link_t *link, char *buf, mtd_request_t *req)
 #ifdef PCMCIA_DEBUG
     time = jiffies - time;
     if (pc_debug && (time > 1))
-	kmsg("write complete, time = %ld, avg = %ld us/word, rate = "
-	     "%ld kb/sec", time, time*20000/req->TransferLength,
-	     req->TransferLength*100/(time*1024));
+	printk(KERN_DEBUG "iflash2_mtd: write complete, time = %ld, "
+	       "avg = %ld us/word, rate = %ld kb/sec\n", time,
+	       time*20000/req->TransferLength,
+	       req->TransferLength*100/(time*1024));
 #endif
     
 done:
@@ -835,9 +829,9 @@ static int flash_erase(dev_link_t *link, char *buf, mtd_request_t *req)
 
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 1)
-	kmsg("flash_erase(0x%p, 0x%lx, 0x%lx, 0x%lx)",
-	     link, req->MediaID, req->DestCardOffset,
-	     req->TransferLength);
+	printk(KERN_DEBUG "iflash2_mtd: flash_erase(0x%p, 0x%lx, "
+	       "0x%lx, 0x%lx)\n", link, req->MediaID,
+	       req->DestCardOffset, req->TransferLength);
 #endif
 
     flash = (flash_region_t *)(req->MediaID);
@@ -892,7 +886,8 @@ static int flash_erase(dev_link_t *link, char *buf, mtd_request_t *req)
 	    goto done;
 	else if (ret != CS_BUSY) {
 	    if (++flash->cell[i].erase_retries > retry_limit) {
-		kmsg("erase failed: too many retries!");
+		printk(KERN_NOTICE "iflash2_mtd: erase failed: "
+		       "too many retries!\n");
 		goto done;
 	    }
 	    else {
@@ -906,7 +901,7 @@ static int flash_erase(dev_link_t *link, char *buf, mtd_request_t *req)
 
     /* If the request is not complete, has it taken too long? */
     if (jiffies > flash->cell[i].erase_time + erase_limit) {
-	kmsg("erase timed out!");
+	printk(KERN_NOTICE "iflash2_mtd: erase timed out!\n");
 	reset_block((u_short *)dev->Base);
 	ret = CS_GENERAL_FAILURE;
 	goto done;
@@ -918,8 +913,8 @@ static int flash_erase(dev_link_t *link, char *buf, mtd_request_t *req)
 done:
 #ifdef PCMCIA_DEBUG
     if (pc_debug)
-	kmsg("erase complete, time = %ld",
-	     jiffies - flash->cell[i].erase_time);
+	printk(KERN_DEBUG "iflash2_mtd: erase complete, time = %ld\n",
+	       jiffies - flash->cell[i].erase_time);
 #endif
     flash->cell[i].state &= ~(FLASH_ERASING|FLASH_PENDING);
     reset_block((u_short *)dev->Base);
@@ -984,7 +979,7 @@ static int flash_event(event_t event, int priority,
 
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 1)
-	kmsg("flash_event(%x)", event);
+	printk(KERN_DEBUG "iflash2_mtd: flash_event(%x)\n", event);
 #endif
     
     switch (event) {
@@ -992,7 +987,7 @@ static int flash_event(event_t event, int priority,
     case CS_EVENT_CARD_REMOVAL:
 	link->state &= ~DEV_PRESENT;
 	if (link->state & DEV_CONFIG) {
-	    link->release.expires = 5;
+	    link->release.expires = RUN_AT(HZ/20);
 	    add_timer(&link->release);
 	}
 	break;
@@ -1024,21 +1019,19 @@ static int flash_event(event_t event, int priority,
 
 /*====================================================================*/
 
-#ifdef MODULE
-char kernel_version[] = UTS_RELEASE;
-
-int init_module(void)
+int init_iflash2_mtd(void)
 {
     servinfo_t serv;
     
 #ifdef PCMCIA_DEBUG
     if (pc_debug)
-	printk(version);
+	printk(KERN_INFO "%s\n", version);
 #endif
     
     CardServices(GetCardServicesInfo, &serv);
     if (serv.Revision != CS_RELEASE_CODE) {
-	kmsg("Card Services release does not match!");
+	printk(KERN_NOTICE "iflash2_mtd: Card Services release "
+	       "does not match!\n");
 	return -1;
     }
     
@@ -1047,9 +1040,13 @@ int init_module(void)
     return 0;
 }
 
+#ifdef MODULE
 void cleanup_module(void)
 {
-    kmsg("unloading");
+#ifdef PCMCIA_DEBUG
+    if (pc_debug)
+	printk(KERN_DEBUG "iflash2_mtd: unloading\n");
+#endif
     unregister_pcmcia_driver(&dev_info);
     while (dev_list != NULL)
 	flash_detach(dev_list);
