@@ -10,11 +10,11 @@
     
 ======================================================================*/
 
-#include <linux/config.h>
+#include "config.h"
+#include "k_compat.h"
 
 #ifdef MODULE
-#include <linux/module.h>
-#include <linux/version.h>
+#define init_pcmem_cs init_module
 #endif
 
 #include <linux/kernel.h>
@@ -36,6 +36,7 @@
 #include "cs.h"
 #include "cistpl.h"
 #include "ds.h"
+#include "mem_op.h"
 
 /* Major device #'s for pcmem device */
 static int major_dev = 0;
@@ -49,12 +50,12 @@ static int major_dev = 0;
 #define DEVICE_NR(device)	(MINOR(device)>>2)
 #define MINOR_NR(dev)		((dev)<<2)
 
-#include <drivers/block/blk.h>
+#include BLK_DEV_HDR
 
 #ifdef PCMCIA_DEBUG
 static int pc_debug = PCMCIA_DEBUG;
 static char *version =
-"pcmem_cs.c 1.43 1995/06/09 20:38:36 (David Hinds)\n";
+"pcmem_cs.c 1.63 1996/05/28 15:16:02 (David Hinds)";
 #endif
 
 /*====================================================================*/
@@ -62,8 +63,8 @@ static char *version =
 /* Maximum number of separate memory devices we'll allow */
 #define MAX_DEV		8
 
-/* Size of the PCMCIA address space: 24 bits */
-#define HIGH_ADDR	0x1000000
+/* Size of the PCMCIA address space: 26 bits = 64 MB */
+#define HIGH_ADDR	0x4000000
 
 /* Size of the memory window: 4K */
 #define WINDOW_SIZE	0x1000
@@ -80,6 +81,7 @@ static dev_link_t *pcmem_attach(void);
 static void pcmem_detach(dev_link_t *);
 
 typedef struct pcmem_dev_t {
+    dev_node_t	node;
     u_long	flags;
     caddr_t	Base;
     u_long	size;
@@ -101,13 +103,15 @@ static int pcmem_ioctl(struct inode *inode, struct file *file,
 		       u_int cmd, u_long arg);
 static int pcmem_read(struct inode *inode, struct file *file,
 		      char *buf, int count);
+static int pcmem_write(struct inode *inode, struct file *file,
+		       CONST char *buf, int count);
 static int pcmem_open(struct inode *inode, struct file *file);
 static void pcmem_close(struct inode *inode, struct file *file);
 
 static struct file_operations pcmem_chr_fops = {
     pcmem_lseek,	/* lseek */
     pcmem_read,		/* read */
-    NULL,		/* write */
+    pcmem_write,	/* write */
     NULL,		/* readdir */
     NULL,		/* select */
     NULL,		/* ioctl */
@@ -132,58 +136,11 @@ static struct file_operations pcmem_blk_fops = {
 
 /*====================================================================*/
 
-static void kmsg(char *fmt, ...)
-{
-    char buf[256];
-    va_list args;
-    
-    va_start(args, fmt);
-    vsprintf(buf, fmt, args);
-    va_end(args);
-    printk("%s: %s\n", (char *)dev_info, buf);
-}
-
 static void cs_error(int func, int ret)
 {
     CardServices(ReportError, dev_info, (void *)func, (void *)ret);
 }
 
-/*======================================================================
-
-    If we're linked into the kernel, this gets called from init.c.
-    
-======================================================================*/
-
-#ifndef MODULE
-u_long pcmem_init(u_long kmem_start, u_long kmem_end)
-{
-    int i;
-    
-#ifdef PCMCIA_DEBUG
-    if (pc_debug)
-	printk(version);
-#endif
-    
-    register_pcmcia_driver(&dev_info, &pcmem_attach, &pcmem_detach);
-    
-    for (i = MAX_CHRDEV-1; i > 0; i--) {
-	if (register_chrdev(i, "pcmem", &pcmem_chr_fops) == 0) {
-	    if (register_blkdev(i, "pcmem", &pcmem_blk_fops) == 0)
-		break;
-	    else
-		unregister_chrdev(i, "pcmem");
-	}
-    }
-    if (i == 0)
-	printk("unable to grab device # for PCMCIA memory driver\n");
-    else
-	major_dev = i;
-    blk_dev[major_dev].request_fn = DEVICE_REQUEST;
-    
-    return kmem_start;
-}
-#endif
-
 /*======================================================================
 
     pcmem_attach() creates an "instance" of the driver, allocating
@@ -201,13 +158,13 @@ static dev_link_t *pcmem_attach(void)
     
 #ifdef PCMCIA_DEBUG
     if (pc_debug)
-	printk("pcmem_attach()\n");
+	printk(KERN_DEBUG "pcmem_attach()\n");
 #endif
 
     for (i = 0; i < MAX_DEV; i++)
 	if (dev_table[i] == NULL) break;
     if (i == MAX_DEV) {
-	kmsg("no devices available");
+	printk(KERN_NOTICE "pcmem_cs: no devices available\n");
 	return NULL;
     }
     
@@ -219,6 +176,7 @@ static dev_link_t *pcmem_attach(void)
     link->release.data = (u_long)link;
     
     dev = kmalloc(sizeof(struct pcmem_dev_t), GFP_KERNEL);
+    memset(dev, 0, sizeof(struct pcmem_dev_t));
     link->priv = dev;
 
     /* Register with Card Services */
@@ -256,7 +214,7 @@ static void pcmem_detach(dev_link_t *link)
 
 #ifdef PCMCIA_DEBUG
     if (pc_debug)
-	printk("pcmem_detach(0x%p)\n", link);
+	printk(KERN_DEBUG "pcmem_detach(0x%p)\n", link);
 #endif
     
     /* Locate device structure */
@@ -266,7 +224,8 @@ static void pcmem_detach(dev_link_t *link)
 	return;
 
     if (link->state & DEV_CONFIG) {
-	kmsg("detach postponed, '%s' still locked", link->dev_name);
+	printk(KERN_INFO "detach postponed, '%s' still locked\n",
+	       link->dev->dev_name);
 	link->state |= DEV_STALE_LINK;
 	return;
     }
@@ -287,50 +246,54 @@ static void pcmem_detach(dev_link_t *link)
     
 ======================================================================*/
 
-#define SIZE_STEP	0x4000
+#define SIZE_STEP	0x10000
 #define SIZE_NSTEP	(HIGH_ADDR/SIZE_STEP)
 
-static u_long get_size(dev_link_t *link, pcmem_dev_t *dev)
+static void get_size(dev_link_t *link, pcmem_dev_t *dev)
 {
     modwin_t mod;
     memreq_t mem;
-    char buf[SIZE_NSTEP];
+    char *buf = kmalloc(SIZE_NSTEP, GFP_KERNEL);
     int s, t, ret;
 
     dev->size = 0;
     mod.Attributes = WIN_ENABLE | WIN_MEMORY_TYPE_CM;
     mod.AccessSpeed = 0;
     ret = CardServices(ModifyWindow, link->win, &mod);
-    if (ret != CS_SUCCESS) {
+    if (ret != CS_SUCCESS)
 	cs_error(ModifyWindow, ret);
-	return 0;
-    }
 
     /* Look for wrap-around or dead end */
     mem.Page = 0;
     for (s = 0; s < SIZE_NSTEP; s++) {
 	mem.CardOffset = s * SIZE_STEP;
 	CardServices(MapMemPage, link->win, &mem);
-	buf[s] = dev->Base[0];
-	dev->Base[0] = ~buf[s];
+	buf[s] = readb(dev->Base);
+	writeb(~buf[s], dev->Base);
 	for (t = 0; t < s; t++) {
 	    mem.CardOffset = t * SIZE_STEP;
 	    CardServices(MapMemPage, link->win, &mem);
-	    if (dev->Base[0] != buf[t]) {
-		dev->Base[0] = buf[t];
+	    if (readb(dev->Base) != buf[t]) {
+		writeb(buf[t], dev->Base);
 		break;
 	    }
 	}
 	if (t < s) break;
 	mem.CardOffset = s * SIZE_STEP;
 	CardServices(MapMemPage, link->win, &mem);
-	if (dev->Base[0] != ~buf[s]) break;
-	dev->Base[0] = buf[s];
+	if (readb(dev->Base) != ~buf[s]) break;
+	writeb(buf[s], dev->Base);
+    }
+
+    /* Restore that last byte on wrap-around */
+    if (t < s) {
+	mem.CardOffset = t * SIZE_STEP;
+	CardServices(MapMemPage, link->win, &mem);
+	writeb(buf[t], dev->Base);
     }
 
     dev->size = s * SIZE_STEP;
-
-    return dev->size;
+    kfree(buf);
 } /* get_size */
 
 /*======================================================================
@@ -350,7 +313,7 @@ static void pcmem_config(dev_link_t *link)
 
 #ifdef PCMCIA_DEBUG
     if (pc_debug)
-	printk("pcmem_config(0x%p)\n", link);
+	printk(KERN_DEBUG "pcmem_config(0x%p)\n", link);
 #endif
 
     /* Configure card */
@@ -385,23 +348,29 @@ static void pcmem_config(dev_link_t *link)
     
     dev = (pcmem_dev_t *)link->priv;
     dev->Base = req.Base;
-    if (status.CardState & CS_EVENT_WRITE_PROTECT)
+    if (status.CardState & CS_EVENT_WRITE_PROTECT) {
 	dev->flags |= MEM_WRPROT;
-    get_size(link, dev);
+	dev->size = 0;
+    }
+    else
+	get_size(link, dev);
 
-    sprintf(link->dev_name, "pcmem%d", i);
-    printk("pcmem: loading ");
+    sprintf(dev->node.dev_name, "pcmem%d", i);
+    printk(KERN_INFO "pcmem_cs: loading ");
     if (dev->size) {
 	if ((dev->size & 0x0fffff) == 0)
 	    printk("%lumb ", dev->size >> 20);
 	else
 	    printk("%luk ", dev->size >> 10);
     }
+    else
+	dev->size = HIGH_ADDR;
     if (dev->flags & MEM_WRPROT)
 	printk("write-protected ");
-    printk("memory card '%s'\n", link->dev_name);
-    link->major = major_dev;
-    link->minor = MINOR_NR(i);
+    printk("memory card '%s'\n", dev->node.dev_name);
+    dev->node.major = major_dev;
+    dev->node.minor = MINOR_NR(i);
+    link->dev = &dev->node;
     link->state &= ~DEV_CONFIG_PENDING;
     
 } /* pcmem_config */
@@ -420,12 +389,15 @@ static void pcmem_release(u_long arg)
 
 #ifdef PCMCIA_DEBUG
     if (pc_debug)
-	printk("pcmem_release(0x%p)\n", link);
+	printk(KERN_DEBUG "pcmem_release(0x%p)\n", link);
 #endif
 
     if (link->open) {
-	printk("pcmem_cs: release postponed, '%s' still open\n",
-	       link->dev_name);
+#ifdef PCMCIA_DEBUG
+	if (pc_debug)
+	    printk(KERN_DEBUG "pcmem_cs: release postponed, '%s' "
+		   "still open\n", link->dev->dev_name);
+#endif
 	link->state |= DEV_STALE_CONFIG;
 	return;
     }
@@ -433,6 +405,7 @@ static void pcmem_release(u_long arg)
     if (link->win)
 	CardServices(ReleaseWindow, link->win);
     link->state &= ~DEV_CONFIG;
+    link->dev = NULL;
     
     if (link->state & DEV_STALE_LINK)
 	pcmem_detach(link);
@@ -455,20 +428,20 @@ static int pcmem_event(event_t event, int priority,
 
 #ifdef PCMCIA_DEBUG
     if (pc_debug)
-	printk("pcmem_event()\n");
+	printk(KERN_DEBUG "pcmem_event()\n");
 #endif
     
     switch (event) {
 #ifdef PCMCIA_DEBUG
     case CS_EVENT_REGISTRATION_COMPLETE:
 	if (pc_debug)
-	    kmsg("registration complete");
+	    printk(KERN_DEBUG "registration complete\n");
 	break;
 #endif
     case CS_EVENT_CARD_REMOVAL:
 	link->state &= ~DEV_PRESENT;
 	if (link->state & DEV_CONFIG) {
-	    link->release.expires = 5;
+	    link->release.expires = RUN_AT(HZ/20);
 	    add_timer(&link->release);
 	}
 	break;
@@ -504,7 +477,7 @@ static int pcmem_open(struct inode *inode, struct file *file)
 
 #ifdef PCMCIA_DEBUG
     if (pc_debug)
-	printk("pcmem_open(%d)\n", MINOR(inode->i_rdev));
+	printk(KERN_DEBUG "pcmem_open(%d)\n", MINOR(inode->i_rdev));
 #endif
 
     link = dev_table[DEVICE_NR(inode->i_rdev)];
@@ -516,9 +489,7 @@ static int pcmem_open(struct inode *inode, struct file *file)
     if ((file->f_mode & 2) && (dev->flags & MEM_WRPROT))
 	return -EROFS;
     
-#ifdef MODULE
     MOD_INC_USE_COUNT;
-#endif
     return 0;
 } /* pcmem_open */
 
@@ -530,15 +501,13 @@ static void pcmem_close(struct inode *inode, struct file *file)
 
 #ifdef PCMCIA_DEBUG
     if (pc_debug)
-	printk("pcmem_close(%d)\n", MINOR(inode->i_rdev));
+	printk(KERN_DEBUG "pcmem_close(%d)\n", MINOR(inode->i_rdev));
 #endif
 
     link = dev_table[DEVICE_NR(inode->i_rdev)];
     if (link == NULL) return;
     
-#ifdef MODULE
     MOD_DEC_USE_COUNT;
-#endif
 
 } /* pcmem_close */
 
@@ -552,23 +521,46 @@ static void pcmem_close(struct inode *inode, struct file *file)
 static int pcmem_lseek(struct inode *inode, struct file *file,
 		       off_t offset, int orig)
 {
+    dev_link_t *link;
+    pcmem_dev_t *dev;
+    
 #ifdef PCMCIA_DEBUG
     int minor = MINOR(inode->i_rdev);
     if (pc_debug)
-	printk("pcmem_lseek(%d, %lu, %d)\n", minor, offset, orig);
+	printk(KERN_DEBUG "pcmem_lseek(%d, %lu, %d)\n",
+	       minor, offset, orig);
 #endif
+    /* Let's just make sure there really is a card here! */
+    link = dev_table[DEVICE_NR(inode->i_rdev)];
+    if (!DEV_OK(link))
+	return -ENODEV;
+
+    dev = (pcmem_dev_t *)link->priv;
+    
     switch (orig) {
     case 0:
 	if ((offset < 0) || (offset > HIGH_ADDR))
 	    return -EINVAL;
+        if (offset > dev->size)
+       	    return -ENOSPC;
 	file->f_pos = offset;
 	break;
     case 1:
 	if ((file->f_pos+offset < 0) ||
 	    (file->f_pos+offset > HIGH_ADDR))
 	    return -EINVAL;
+	if (file->f_pos+offset > dev->size)
+	    return -ENOSPC;
 	file->f_pos += offset;
 	break;
+    case 2:
+	if (((int)dev->size+offset < 0) ||
+	    (dev->size+offset > HIGH_ADDR))
+	    return -EINVAL;
+        if (dev->size+offset > dev->size)
+       	    return -ENOSPC;
+	file->f_pos = dev->size+offset;
+	break;
     default:
 	return -EINVAL;
     }
@@ -594,7 +586,7 @@ static int pcmem_read(struct inode *inode, struct file *file,
 
 #ifdef PCMCIA_DEBUG
     if (pc_debug)
-	printk("pcmem_read(%d, %d)\n", minor, count);
+	printk(KERN_DEBUG "pcmem_read(%d, %d)\n", minor, count);
 #endif
 
     link = dev_table[DEVICE_NR(inode->i_rdev)];
@@ -607,9 +599,12 @@ static int pcmem_read(struct inode *inode, struct file *file,
     if (count < 0)
 	return -EINVAL;
     if (p >= HIGH_ADDR)
-	return 0;
-    if (count > HIGH_ADDR - p)
-	count = HIGH_ADDR - p;
+	return -EFBIG;
+    if (p >= dev->size)
+	return -ENOSPC;
+    /* Don't read more than # of bytes left */
+    if (count > dev->size - p)
+	count = dev->size - p;
 
     mod.Attributes = WIN_ENABLE;
     mod.AccessSpeed = 0;
@@ -636,12 +631,12 @@ static int pcmem_read(struct inode *inode, struct file *file,
 	nb = (from+count > WINDOW_SIZE) ? WINDOW_SIZE-from : count;
 	if (minor & 1) {
 	    for (i = 0; i < nb; i += 2, buf++)
-		put_fs_byte(dev->Base[from+i], buf);
+		put_fs_byte(readb(dev->Base+from+i), buf);
 	    read += nb>>1;
 	}
 	else {
-	    for (i = 0; i < nb; i++, buf++)
-		put_fs_byte(dev->Base[from+i], buf);
+	    memcpy_frompc_tofs(buf, dev->Base+from, nb);
+	    buf += nb;
 	    read += nb;
 	}
         from = 0;
@@ -652,6 +647,90 @@ static int pcmem_read(struct inode *inode, struct file *file,
     return read;
 } /* pcmem_read */
 
+/*======================================================================
+
+    Write for character-mode device
+    
+======================================================================*/
+
+static int pcmem_write(struct inode *inode, struct file *file,
+		       CONST char *buf, int count)
+{
+    int minor = MINOR(inode->i_rdev);
+    dev_link_t *link;
+    pcmem_dev_t *dev;
+    u_long p, wrote, to, nb;
+    int i, ret;
+    modwin_t mod;
+    memreq_t mem;
+    
+#ifdef PCMCIA_DEBUG
+    if (pc_debug)
+	printk(KERN_DEBUG "pcmem_write(%d, %d)\n", minor, count);
+#endif
+    
+    link = dev_table[DEVICE_NR(inode->i_rdev)];
+    
+    if (!DEV_OK(link))
+	return -ENODEV;
+    
+    dev = (pcmem_dev_t *)link->priv;
+    
+    /* Check for write protect */
+    if (dev->flags & MEM_WRPROT)
+	return -EROFS;
+    
+    p = file->f_pos;
+    if (count < 0)
+	return -EINVAL;
+    if (p >= HIGH_ADDR)
+	return -EFBIG;
+    if (p >= dev->size)
+        return -ENOSPC;
+    if (count > dev->size - p)
+	count = dev->size - p;
+
+    mod.Attributes = WIN_ENABLE;
+    mod.AccessSpeed = 0;
+    if (minor & 1) {
+	mod.Attributes |= WIN_MEMORY_TYPE_AM;
+	p = p * 2; count = count * 2;
+    }
+    ret = CardServices(ModifyWindow, link->win, &mod);
+    if (ret != CS_SUCCESS) {
+	cs_error(ModifyWindow, ret);
+	return -EIO;
+    }
+    
+    mem.CardOffset = p & ~(WINDOW_SIZE-1);
+    mem.Page = 0;
+    to = p & (WINDOW_SIZE-1);
+    wrote = 0;
+    for ( ; count > 0; count -= nb) {
+	ret = CardServices(MapMemPage, link->win, &mem);
+	if (ret != CS_SUCCESS) {
+	    cs_error(MapMemPage, ret);
+	    return -EIO;
+	}
+	nb = (to+count > WINDOW_SIZE) ? WINDOW_SIZE-to : count;
+	if (minor & 1) {
+	    for (i = 0; i < nb; i += 2, buf++)
+		writeb(get_fs_byte(buf), dev->Base+to+i);
+	    wrote += nb>>1;
+	}
+	else {
+	    memcpy_fromfs_topc(dev->Base+to, buf, nb);
+	    buf += nb;
+	    wrote += nb;
+	}
+        to = 0;
+	mem.CardOffset += WINDOW_SIZE;
+    }
+
+    file->f_pos += wrote;
+    return wrote;
+} /* pcmem_write */
+
 /*======================================================================
 
     IOCTL calls for getting device parameters.
@@ -705,7 +784,7 @@ static void do_pcmem_request(void)
     do {
 	INIT_REQUEST;
 
-	link = dev_table[DEVICE_NR(CURRENT->dev)];
+	link = dev_table[DEVICE_NR(DEVICE(CURRENT))];
 	
 	if (!DEV_OK(link)) {
 	    end_request(0);
@@ -738,7 +817,7 @@ static void do_pcmem_request(void)
 		ret = CardServices(MapMemPage, link->win, &mem);
 		if (ret != CS_SUCCESS) break;
 		nb = (from+len > WINDOW_SIZE) ? WINDOW_SIZE-from : len;
-		memcpy(buf, &dev->Base[from], nb);
+		memcpy_frompc(buf, &dev->Base[from], nb);
 		mem.CardOffset += WINDOW_SIZE;
 	    }
 	    break;
@@ -748,7 +827,7 @@ static void do_pcmem_request(void)
 		ret = CardServices(MapMemPage, link->win, &mem);
 		if (ret != CS_SUCCESS) break;
 		nb = (from+len > WINDOW_SIZE) ? WINDOW_SIZE-from : len;
-		memcpy(&dev->Base[from], buf, nb);
+		memcpy_topc(&dev->Base[from], buf, nb);
 		mem.CardOffset += WINDOW_SIZE;
 	    }
 	    break;
@@ -768,22 +847,20 @@ static void do_pcmem_request(void)
 
 /*====================================================================*/
 
-#ifdef MODULE
-char kernel_version[] = UTS_RELEASE;
-
-int init_module(void)
+int init_pcmem_cs(void)
 {
     servinfo_t serv;
     int i;
     
 #ifdef PCMCIA_DEBUG
     if (pc_debug)
-	printk(version);
+	printk(KERN_INFO "%s\n", version);
 #endif
     
     CardServices(GetCardServicesInfo, &serv);
     if (serv.Revision != CS_RELEASE_CODE) {
-	printk("pcmem_cs: Card Services release does not match!\n");
+	printk(KERN_NOTICE "pcmem_cs: Card Services release "
+	       "does not match!\n");
 	return -1;
     }
     
@@ -798,7 +875,7 @@ int init_module(void)
 	}
     }
     if (i == 0)
-	printk("unable to grab device # for PCMCIA memory driver\n");
+	printk(KERN_NOTICE "pcmem_cs: unable to grab a device #\n");
     else
 	major_dev = i;
 
@@ -810,12 +887,16 @@ int init_module(void)
     return 0;
 }
 
+#ifdef MODULE
 void cleanup_module(void)
 {
     int i;
     dev_link_t *link;
-    
-    kmsg("unloading");
+
+#ifdef PCMCIA_DEBUG
+    if (pc_debug)
+	printk(KERN_DEBUG "pcmem_cs: unloading\n");
+#endif
     unregister_pcmcia_driver(&dev_info);
     if (major_dev != 0) {
 	unregister_chrdev(major_dev, "pcmem");
