@@ -73,7 +73,7 @@ static char *if_names[] = { "Auto", "10baseT", "10base2"};
 #ifdef PCMCIA_DEBUG
 static int pc_debug = PCMCIA_DEBUG;
 static char *version =
-"pcnet_cs.c 1.37 1996/06/04 15:50:09 (David Hinds)";
+"pcnet_cs.c 1.41 1996/08/06 03:13:59 (David Hinds)";
 #endif
 
 /*====================================================================*/
@@ -166,17 +166,21 @@ static hw_info_t hw_info[] = {
 	  USE_SHMEM | HAS_MISC_REG | HAS_IBM_MISC },
     { "IBM FME", 0x0374, 0x00, 0x04, 0xac,
           USE_SHMEM | HAS_MISC_REG | HAS_IBM_MISC },
+    { "Hypertec Ethernet", 0x01c0, 0x00, 0x40, 0x4c, 0 },
     { "Kingston KNE-PCM/x", 0x0ff0, 0x00, 0xc0, 0xf0,
 	  USE_SHMEM | HAS_MISC_REG | HAS_IBM_MISC },
     { "Kingston KNE-PCM/x", 0x0ff0, 0xe2, 0x0c, 0x0f,
 	  USE_SHMEM | HAS_MISC_REG | HAS_IBM_MISC },
+    { "Network General Sniffer", 0x0ff0, 0x00, 0x00, 0x65,
+	  USE_SHMEM | HAS_MISC_REG | HAS_IBM_MISC },
     { "Grey Cell GCS2220", 0x0000, 0x00, 0x47, 0x43, 0 },
     { "Allied Telesis LA-PCM", 0x0ff0, 0x00, 0x00, 0xf4, 0 },
     { "SCM Ethernet", 0x0ff0, 0x00, 0x20, 0xcb, 0 },
     { "NDC Instant-Link", 0x003a, 0x00, 0x80, 0xc6, 0 },
     { "Epson EEN10B", 0x0ff0, 0x00, 0x00, 0x48,
           USE_SHMEM | HAS_MISC_REG | HAS_IBM_MISC },
-    { "Longshine LCS-8534", 0, 0x08, 0x00, 0x00, 0 }
+    { "Longshine LCS-8534", 0, 0x08, 0x00, 0x00, 0 },
+    { "ELECOM Laneed LD-CDWA", 0xb8, 0x08, 0x00, 0x42, 0 }
 };
 
 #define NR_INFO		(sizeof(hw_info)/sizeof(hw_info_t))
@@ -519,7 +523,7 @@ static void pcnet_config(dev_link_t *link)
     CS_EXIT_TEST(i, GetTupleData, config_failed);
     i = CardServices(ParseTuple, handle, &tuple, &parse);
     CS_EXIT_TEST(i, ParseTuple, config_failed);
-    link->conf.ConfigBase = (caddr_t)parse.config.base;
+    link->conf.ConfigBase = parse.config.base;
     link->conf.Present = parse.config.rmask[0];
 
     /* Configure card */
@@ -895,7 +899,7 @@ static void pcnet_reset_8390(struct device *dev)
 
 static int set_config(struct device *dev, struct ifmap *map)
 {
-    if (map->port != (u_char)(-1)) {
+    if ((map->port != (u_char)(-1)) && (map->port != dev->if_port)) {
 	if ((map->port == 1) || (map->port == 2)) {
 	    dev->if_port = map->port;
 	    printk(KERN_INFO "%s: switched to %s port\n",
