@@ -38,7 +38,7 @@
 #ifdef PCMCIA_DEBUG
 static int pc_debug = PCMCIA_DEBUG;
 static char *version =
-"skeleton.c 1.30 1996/05/28 15:25:32 (David Hinds)";
+"skeleton.c 1.32 1996/08/17 15:19:54 (David Hinds)";
 #endif
 
 /*====================================================================*/
@@ -60,7 +60,7 @@ static u_long irq_mask = 0xdeb8;
    handler.
 */
 
-static void skeleton_config(u_long arg);
+static void skeleton_config(dev_link_t *link);
 static void skeleton_release(u_long arg);
 static int skeleton_event(event_t event, int priority,
 			  event_callback_args_t *args);
@@ -121,7 +121,7 @@ static dev_link_t *dev_list = NULL;
 */
    
 typedef struct local_info_t {
-    dev_node_t	*node;
+    dev_node_t	node;
 } local_info_t;
 
 /*====================================================================*/
@@ -273,7 +273,7 @@ static void skeleton_config(dev_link_t *link)
     client_handle_t handle;
     tuple_t tuple;
     cisparse_t parse;
-    struct device *dev;
+    local_info_t *dev;
     int i, j;
     u_char buf[64];
     win_req_t req;
@@ -303,7 +303,7 @@ static void skeleton_config(dev_link_t *link)
 	if (i != CS_SUCCESS) break;
 	i = CardServices(ParseTuple, handle, &tuple, &parse);
 	if (i != CS_SUCCESS) break;
-	link->conf.ConfigBase = (caddr_t)parse.config.base;
+	link->conf.ConfigBase = parse.config.base;
     } while (0);
     if (i != CS_SUCCESS) {
 	cs_error(ParseTuple, i);
@@ -365,7 +365,7 @@ static void skeleton_config(dev_link_t *link)
 	link->win = (window_handle_t)link->handle;
 	i = CardServices(RequestWindow, &link->win, &req);
 	if (i != 0) {
-	    cs_error(RequestWindow, ret);
+	    cs_error(RequestWindow, i);
 	    break;
 	}
 
