@@ -2,9 +2,8 @@
 
     A PCMCIA ethernet driver for SMC91c92-based cards.
 
-    This driver supports the Megahertz PCMCIA ethernet cards, the SMC
-    ethernet cards, and the Ositech ethernet/modem multifunction
-    cards.
+    This driver supports Megahertz PCMCIA ethernet cards, and Ositech
+    ethernet/modem multifunction cards.
 
     This driver contains code written by Donald Becker
     (becker@cesdis.gsfc.nasa.gov), Rowan Hughes (x-csrdh@jcu.edu.au),
@@ -99,12 +98,16 @@ static dev_link_t *dev_list = NULL;
 
 struct smc_private {
     u_short			manfid;
+    u_short			cardid;
     struct enet_statistics	stats;
     dev_node_t			node;
     struct sk_buff		*saved_skb;
     int				packets_waiting;
 };
 
+/* Special definitions for Megahertz multifunction cards */
+#define MEGAHERTZ_ISR		0x0380
+
 /* Special function registers for Ositech cards */
 #define OSITECH_AUI_CTL		0x0c
 #define OSITECH_PWRDOWN		0x0d
@@ -141,7 +144,7 @@ struct smc_private {
 #define  CFG_DIS_LINK		0x0040
 #define  CFG_STATIC		0x0030
 #define  CFG_IRQ_SEL_1		0x0004
-#define  CFG_IRQ_SEL_0		0x0004
+#define  CFG_IRQ_SEL_0		0x0002
 #define	ADDR0			4
 #define	CONTROL			12
 #define  CTL_AUTO_RELEASE	0x0800
@@ -423,6 +426,9 @@ static int mhz_setup(dev_link_t *link)
     tuple.TupleOffset = 0;
     if (CardServices(GetFirstTuple, handle, &tuple) != CS_SUCCESS)
 	return -1;
+    /* Ugh -- the EM1144 card has two VERS_1 tuples!?! */
+    if (CardServices(GetNextTuple, handle, &tuple) != CS_SUCCESS)
+	CardServices(GetFirstTuple, handle, &tuple);
     if (CardServices(GetTupleData, handle, &tuple) != CS_SUCCESS)
 	return -1;
     if (CardServices(ParseTuple, handle, &tuple, &parse) != CS_SUCCESS)
@@ -431,7 +437,7 @@ static int mhz_setup(dev_link_t *link)
     if (cvt_ascii_address(dev, station_addr) == 0)
 	return 0;
 
-    /* Another possibility: in a special tuple */
+    /* Another possibility: for the EM3288, in a special tuple */
     tuple.DesiredTuple = 0x81;
     if (CardServices(GetFirstTuple, handle, &tuple) != CS_SUCCESS)
 	return -1;
@@ -444,56 +450,48 @@ static int mhz_setup(dev_link_t *link)
     return -1;
 }
 
-/*====================================================================*/
-
-static int smc_setup(dev_link_t *link)
+static void busy_loop(u_long len)
 {
-    client_handle_t handle = link->handle;
-    struct device *dev = link->priv;
-    tuple_t tuple;
-    cisparse_t parse;
-    cistpl_lan_node_id_t *node_id;
-    u_char buf[255], *station_addr;
-    int i;
-    
-    /* Read the station address from the CIS.  It is stored in the
-       third string in the Version 1 Version/ID tuple. */
-    tuple.Attributes = 0;
-    tuple.DesiredTuple = CISTPL_VERS_1;
-    tuple.TupleData = buf;
-    tuple.TupleDataMax = sizeof(buf);
-    tuple.TupleOffset = 0;
-    if (CardServices(GetFirstTuple, handle, &tuple) != CS_SUCCESS)
-	return -1;
-    if (CardServices(GetTupleData, handle, &tuple) != CS_SUCCESS)
-	return -1;
-    if (CardServices(ParseTuple, handle, &tuple, &parse) != CS_SUCCESS)
-	return -1;
-    station_addr = parse.version_1.str + parse.version_1.ofs[2];
-    if (cvt_ascii_address(dev, station_addr) == 0)
-	return 0;
+    u_long timeout = jiffies + len;
+    u_long flags;
+    save_flags(flags);
+    sti();
+    while (timeout >= jiffies) ;
+    restore_flags(flags);
+}
 
-    /* Another possibility: in the LAN function extension tuples */
-    tuple.DesiredTuple = CISTPL_FUNCE;
-    if (CardServices(GetFirstTuple, handle, &tuple) != CS_SUCCESS)
+static int mhz_3288_power(dev_link_t *link)
+{
+    win_req_t req;
+    memreq_t mem;
+    volatile u_char *attr;
+    u_char tmp;
+    
+    req.Attributes = WIN_DATA_WIDTH_8|WIN_MEMORY_TYPE_AM|WIN_ENABLE;
+    req.Base = NULL;
+    req.Size = 0x1000;
+    req.AccessSpeed = 0;
+    link->win = (window_handle_t)link->handle;
+    if (CardServices(RequestWindow, &link->win, &req) != CS_SUCCESS)
 	return -1;
-    for (;;) {
-	if (CardServices(GetTupleData, handle, &tuple) != CS_SUCCESS)
-	    return -1;
-	if (CardServices(ParseTuple, handle, &tuple, &parse) != CS_SUCCESS)
-	    return -1;
-	if (parse.funce.type == CISTPL_FUNCE_LAN_NODE_ID)
-	    break;
-	if (CardServices(GetNextTuple, handle, &tuple) != CS_SUCCESS)
-	    return -1;
-    }
-    node_id = (cistpl_lan_node_id_t *)parse.funce.data;
-    if (node_id->nb == 6) {
-	for (i = 0; i < 6; i++)
-	    dev->dev_addr[i] = node_id->id[i];
-	return 0;
-    }
-    return -1;
+    mem.Page = mem.CardOffset = 0;
+    if (CardServices(MapMemPage, link->win, &mem) != CS_SUCCESS)
+	return -1;
+
+    /* Read the ISR twice... */
+    attr = req.Base;
+    tmp = attr[MEGAHERTZ_ISR];
+    udelay(5);
+    tmp = attr[MEGAHERTZ_ISR];
+    /* Pause 200ms... */
+    busy_loop(HZ/5);
+    /* Now read and write the COR... */
+    tmp = attr[link->conf.ConfigBase + CISREG_COR];
+    udelay(5);
+    attr[link->conf.ConfigBase + CISREG_COR] = tmp;
+
+    CardServices(ReleaseWindow, link->win);
+    return 0;
 }
 
 /*====================================================================*/
@@ -551,7 +549,8 @@ static void smc91c92_config(dev_link_t *link)
     tuple_t tuple;
     cisparse_t parse;
     u_short ioaddr, buf[32];
-    int i, j;
+    int i, j, k;
+    static u_short com[4] = { 0x03f8, 0x2f8, 0x3e8, 0x2e8 };
 
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 2)
@@ -570,7 +569,7 @@ static void smc91c92_config(dev_link_t *link)
     CS_EXIT_TEST(i, GetTupleData, config_failed);
     i = CardServices(ParseTuple, handle, &tuple, &parse);
     CS_EXIT_TEST(i, ParseTuple, config_failed);
-    link->conf.ConfigBase = (caddr_t)parse.config.base;
+    link->conf.ConfigBase = parse.config.base;
     link->conf.Present = parse.config.rmask[0];
     
     tuple.DesiredTuple = CISTPL_MANFID;
@@ -579,6 +578,7 @@ static void smc91c92_config(dev_link_t *link)
     i = CardServices(GetTupleData, handle, &tuple);
     CS_EXIT_TEST(i, GetTupleData, config_failed);
     lp->manfid = buf[0];
+    lp->cardid = buf[1];
     
     /* Configure card */
     link->state |= DEV_CONFIG;
@@ -594,7 +594,6 @@ static void smc91c92_config(dev_link_t *link)
 	/* Enable Hard Decode, LAN, Modem */
 	link->conf.ConfigIndex = 0x23;
 	for (j = 0; j < 4; j++) {
-	    static u_short com[4] = { 0x03f8, 0x2f8, 0x3e8, 0x2e8 };
 	    link->io.BasePort2 = com[j];
 	    i = CardServices(RequestIO, link->handle, &link->io);
 	    if (i == CS_SUCCESS) break;
@@ -606,10 +605,41 @@ static void smc91c92_config(dev_link_t *link)
 	    i = CardServices(RequestIO, link->handle, &link->io);
 	}
     }
+    else if ((lp->manfid == MANFID_MEGAHERTZ) &&
+	     ((lp->cardid == PRODID_MEGAHERTZ_VARIOUS) ||
+	      (lp->cardid == PRODID_MEGAHERTZ_EM3288))) {
+	cistpl_cftable_entry_t *cf = &parse.cftable_entry;
+	
+	link->conf.Attributes |= CONF_ENABLE_SPKR;
+	link->conf.Status = CCSR_AUDIO_ENA;
+	link->irq.Attributes =
+	    IRQ_TYPE_DYNAMIC_SHARING|IRQ_FIRST_SHARED|IRQ_HANDLE_PRESENT;
+	link->io.Attributes2 = IO_DATA_PATH_WIDTH_8;
+	link->io.NumPorts2 = 8;
+	tuple.DesiredTuple = CISTPL_CFTABLE_ENTRY;
+	i = CardServices(GetFirstTuple, handle, &tuple);
+	/* The Megahertz combo cards have modem-like CIS entries, so
+	   we have to explicitly try a bunch of port combinations. */
+	for (j = 0; j < 4; j++) {
+	    i = CardServices(GetTupleData, handle, &tuple);
+	    if (i != CS_SUCCESS) break;
+	    i = CardServices(ParseTuple, handle, &tuple, &parse);
+	    if (i != CS_SUCCESS) break;
+	    link->conf.ConfigIndex = cf->index;
+	    link->io.BasePort2 = cf->io.win[0].base;
+	    for (k = 0x300; k < 0x380; k += 0x20) {
+		link->io.BasePort1 = k;
+		i = CardServices(RequestIO, link->handle, &link->io);
+		if (i == CS_SUCCESS) break;
+	    }
+	    if (i == CS_SUCCESS) break;
+	    i = CardServices(GetNextTuple, handle, &tuple);
+	}
+    }
     else {
 	link->io.NumPorts1 = 16;
 	link->conf.ConfigIndex = 0x01;
-	for (j = 0x300; j < 0x400; j += 0x10) {
+	for (j = 0x200; j < 0x400; j += 0x10) {
 	    link->io.BasePort1 = j;
 	    i = CardServices(RequestIO, link->handle, &link->io);
 	    if (i == CS_SUCCESS) break;
@@ -617,6 +647,10 @@ static void smc91c92_config(dev_link_t *link)
     }
     CS_EXIT_TEST(i, RequestIO, config_failed);
     
+    if ((lp->manfid == MANFID_MEGAHERTZ) &&
+	(lp->cardid == PRODID_MEGAHERTZ_EM3288))
+	mhz_3288_power(link);
+    
     i = CardServices(RequestIRQ, link->handle, &link->irq);
     CS_EXIT_TEST(i, RequestIRQ, config_failed);
     i = CardServices(RequestConfiguration, link->handle, &link->conf);
@@ -627,7 +661,7 @@ static void smc91c92_config(dev_link_t *link)
 	ioaddr = dev->base_addr = link->io.BasePort1 + 0x10;
     else
 	ioaddr = dev->base_addr = link->io.BasePort1;
-    
+
     if ((if_port == 1) || (if_port == 2))
 	dev->if_port = if_port;
     else
@@ -642,9 +676,7 @@ static void smc91c92_config(dev_link_t *link)
     switch (lp->manfid) {
     case MANFID_OSITECH:
 	i = osi_setup(link); break;
-    case MANFID_SMC:
-	i = smc_setup(link); break;
-    case 0x0128: /* Ugly, but some Megahertz cards are weird... */
+    case 0x128: /* For broken Megahertz cards */
     case MANFID_MEGAHERTZ:
 	i = mhz_setup(link); break;
     default:
@@ -652,16 +684,6 @@ static void smc91c92_config(dev_link_t *link)
 	goto config_undo;
     }
     
-#if 0
-    SMC_SELECT_BANK( 1 );
-    for ( i = 0; i < 6; i += 2 ) { 
-	uint16 address;
-	address = inw(ioaddr + ADDR0 + i);
-	dev->dev_addr[i+1] = address >> 8;
-	dev->dev_addr[i] = address & 0xFF;
-    }
-#endif
-    
     if (i != 0) {
 	printk(KERN_NOTICE "smc91c92_cs: Unable to find hardware address.\n");
 	link->state &= ~DEV_CONFIG_PENDING;
@@ -671,7 +693,7 @@ static void smc91c92_config(dev_link_t *link)
     link->dev = &lp->node;
     link->state &= ~DEV_CONFIG_PENDING;
     
-    printk(KERN_INFO "%s: SMC91c92 at port %#3lx, irq %d, %s port ",
+    printk(KERN_INFO "%s: SMC91c92 at port %#3lx, irq %d, %s port, ",
 	   dev->name, dev->base_addr, dev->irq, if_names[dev->if_port]);
     for (i = 0; i < 5; i++)
 	printk("%02X:", dev->dev_addr[i]);
@@ -744,7 +766,7 @@ static int smc91c92_event(event_t event, int priority,
 {
     dev_link_t *link = args->client_data;
     struct device *dev = link->priv;
-
+    
 #ifdef PCMCIA_DEBUG
     if (pc_debug)
 	printk(KERN_DEBUG "smc91c92_event()\n");
@@ -787,11 +809,19 @@ static int smc91c92_event(event_t event, int priority,
 	/* Fall through... */
     case CS_EVENT_CARD_RESET:
 	if (link->state & DEV_CONFIG) {
+	    struct smc_private *lp = dev->priv;
+	    if ((lp->manfid == MANFID_MEGAHERTZ) &&
+		(lp->cardid == PRODID_MEGAHERTZ_EM3288))
+		mhz_3288_power(link);
 	    CardServices(RequestConfiguration, link->handle, &link->conf);
+	    if (lp->manfid == MANFID_OSITECH) {
+		/* Power up the card and enable interrupts */
+		outb(0x03, dev->base_addr - 0x10 + OSITECH_PWRDOWN);
+		outb(0x03, dev->base_addr - 0x10 + OSITECH_ISR);
+	    }
 	    if (link->open) {
 		smc_reset(dev);
-		dev->tbusy = 0;
-				dev->start = 1;
+		dev->tbusy = 0;	dev->start = 1;
 	    }
 	}
 	break;
@@ -806,16 +836,6 @@ static int smc91c92_event(event_t event, int priority,
     
 ======================================================================*/
 
-static void busy_loop(u_long len)
-{
-    u_long timeout = jiffies + len;
-    u_long flags;
-    save_flags(flags);
-    sti();
-    while (timeout >= jiffies) ;
-    restore_flags(flags);
-}
-
 static int smc91c92_open(struct device *dev)
 {
     struct smc_private *lp = (struct smc_private *)dev->priv;
@@ -1202,6 +1222,7 @@ static void smc_interrupt IRQ(int irq, void *instance, struct pt_regs *regs)
 	       dev->name, irq, ioaddr);
 #endif
     
+    lp = (struct smc_private *)dev->priv;
     saved_bank = inw(ioaddr + BANK_SELECT);
     if ((saved_bank & 0xff00) != 0x3300) {
 	/* The device does not exist -- the card could be off-line, or
@@ -1212,11 +1233,15 @@ static void smc_interrupt IRQ(int irq, void *instance, struct pt_regs *regs)
 		   "/ejected device.\n", dev->name, irq);
 	dev->interrupt = 0;
 #endif
+	if (lp->manfid == MANFID_OSITECH) {
+	    /* Retrigger interrupt if needed */
+	    outb(0x00, ioaddr - 0x10 + OSITECH_ISR);
+	    outb(0x03, ioaddr - 0x10 + OSITECH_ISR);
+	}
 	return;
     }
     
     SMC_SELECT_BANK(2);
-    lp = (struct smc_private *)dev->priv;
     saved_pointer = inw(ioaddr + POINTER);
     mask = inb(ioaddr + INT_MASK);
     /* clear all interrupts */
@@ -1286,8 +1311,12 @@ static void smc_interrupt IRQ(int irq, void *instance, struct pt_regs *regs)
 	printk(KERN_DEBUG "%s: Exiting interrupt IRQ%d.\n",
 	       dev->name, irq);
 #endif
-    
-    return;
+
+    if (lp->manfid == MANFID_OSITECH) {
+	/* Retrigger interrupt if needed */
+	outb(0x00, ioaddr - 0x10 + OSITECH_ISR);
+	outb(0x03, ioaddr - 0x10 + OSITECH_ISR);
+    }
 }
 
 /*====================================================================*/
@@ -1471,7 +1500,7 @@ static void set_rx_mode(struct device *dev)
 
 static int s9k_config(struct device *dev, struct ifmap *map)
 {
-    if (map->port != (u_char)(-1)) {
+    if ((map->port != (u_char)(-1)) && (map->port != dev->if_port)) {
 	if ((map->port == 1) || (map->port == 2)) {
 	    dev->if_port = map->port;
 	    printk(KERN_INFO "%s: switched to %s port\n",
