@@ -6,10 +6,11 @@
     
 ======================================================================*/
 
+#include "config.h"
+#include "k_compat.h"
+
 #ifdef MODULE
-#include <linux/autoconf.h>
-#include <linux/module.h>
-#include <linux/version.h>
+#define init_tcic init_module
 #endif
 
 #include <sys/types.h>
@@ -37,7 +38,7 @@
 #ifdef PCMCIA_DEBUG
 static int pc_debug = PCMCIA_DEBUG;
 static const char *version =
-    "tcic.c 1.54 1995/07/06 19:39:18 (David Hinds)\n";
+"tcic.c 1.75 1996/08/29 04:43:49 (David Hinds)";
 #endif
 
 /*====================================================================*/
@@ -50,8 +51,11 @@ static int tcic_base = TCIC_BASE;
 /* Specify a socket number to ignore */
 static int ignore = -1;
 
-/* Bit map of interrupts to choose from -- 0 means just autoprobe */
-static u_long irq_mask = 0;
+/* Probe for safe interrupts? */
+static int do_scan = 1;
+
+/* Bit map of interrupts to choose from */
+static u_long irq_mask = 0xffff;
 
 /* The card status change interrupt -- 0 means autoselect */
 static int cs_irq = 0;
@@ -60,7 +64,7 @@ static int cs_irq = 0;
 static int poll_interval = 0;
 
 /* Delay for card status double-checking */
-static int poll_quick = 5;
+static int poll_quick = HZ/20;
 
 /* CCLK external clock time, in nanoseconds.  70 ns = 14.31818 MHz */
 static int cycle_time = 70;
@@ -70,7 +74,7 @@ static int async_clock = 0;
 
 /*====================================================================*/
 
-static void tcic_interrupt(int irq, struct pt_regs *regs);
+static void tcic_interrupt IRQ(int irq, void *dev, struct pt_regs *regs);
 static void tcic_timer(u_long data);
 static int tcic_service(u_int sock, u_int cmd, void *arg);
 
@@ -79,6 +83,7 @@ typedef struct socket_info_t {
     void	(*handler)(void *info, u_long events);
     void	*info;
     u_char	last_sstat;
+    u_char	id;
 } socket_info_t;
 
 static struct timer_list poll_timer;
@@ -97,26 +102,26 @@ static socket_cap_t tcic_cap = {
 static u_char tcic_getb(u_char reg)
 {
     u_char val = inb(tcic_base+reg);
-    printk("tcic_getb(%#x) = %#x\n", tcic_base+reg, val);
+    printk(KERN_DEBUG "tcic_getb(%#x) = %#x\n", tcic_base+reg, val);
     return val;
 }
 
 static u_short tcic_getw(u_char reg)
 {
     u_short val = inw(tcic_base+reg);
-    printk("tcic_getw(%#x) = %#x\n", tcic_base+reg, val);
+    printk(KERN_DEBUG "tcic_getw(%#x) = %#x\n", tcic_base+reg, val);
     return val;
 }
 
 static void tcic_setb(u_char reg, u_char data)
 {
-    printk("tcic_setb(%#x, %#x)\n", tcic_base+reg, data);
+    printk(KERN_DEBUG "tcic_setb(%#x, %#x)\n", tcic_base+reg, data);
     outb(data, tcic_base+reg);
 }
 
 static void tcic_setw(u_char reg, u_short data)
 {
-    printk("tcic_setw(%#x, %#x)\n", tcic_base+reg, data);
+    printk(KERN_DEBUG "tcic_setw(%#x, %#x)\n", tcic_base+reg, data);
     outw(data, tcic_base+reg);
 }
 #else
@@ -129,7 +134,7 @@ static void tcic_setw(u_char reg, u_short data)
 static void tcic_setl(u_char reg, u_long data)
 {
 #ifdef PCMCIA_DEBUG_X
-    printk("tcic_setl(%#x, %#lx)\n", tcic_base+reg, data);
+    printk(KERN_DEBUG "tcic_setl(%#x, %#lx)\n", tcic_base+reg, data);
 #endif
     outw(data & 0xffff, tcic_base+reg);
     outw(data >> 16, tcic_base+reg+2);
@@ -149,14 +154,12 @@ static void tcic_aux_setb(u_short reg, u_char data)
     tcic_setb(TCIC_AUX, data);
 }
 
-#if 0
 static u_short tcic_aux_getw(u_short reg)
 {
     u_char mode = (tcic_getb(TCIC_MODE) & TCIC_MODE_PGMMASK) | reg;
     tcic_setb(TCIC_MODE, mode);
     return tcic_getw(TCIC_AUX);
 }
-#endif
 
 static void tcic_aux_setw(u_short reg, u_short data)
 {
@@ -198,7 +201,7 @@ static void busy_loop(u_long len)
 
 static volatile u_long irq_hits;
 
-static void irq_count(int irq, struct pt_regs *regs)
+static void irq_count IRQ(int irq, void *dev, struct pt_regs *regs)
 {
     irq_hits++;
 }
@@ -207,12 +210,12 @@ static u_long try_irq(int irq)
 {
     u_short cfg;
 
-    if (request_irq(irq, irq_count, 0, "irq scan") != 0)
+    if (REQUEST_IRQ(irq, irq_count, 0, "irq scan") != 0)
 	return -1;
     irq_hits = 0;
-    busy_loop(2);
+    busy_loop(HZ/50);
     if (irq_hits) {
-	free_irq(irq);
+	FREE_IRQ(irq);
 	return -1;
     }
 
@@ -222,8 +225,8 @@ static u_long try_irq(int irq)
     tcic_setb(TCIC_IENA, TCIC_IENA_ERR | TCIC_IENA_CFG_HIGH);
     tcic_setb(TCIC_ICSR, TCIC_ICSR_ERR | TCIC_ICSR_JAM);
 
-    busy_loop(2);
-    free_irq(irq);
+    busy_loop(HZ/50);
+    FREE_IRQ(irq);
 
     /* Turn off interrupts */
     tcic_setb(TCIC_IENA, TCIC_IENA_CFG_OFF);
@@ -239,6 +242,14 @@ static u_long irq_scan(u_long mask0)
     u_long mask1, mask2;
     int i, ni;
 
+#ifdef __alpha__
+#define PIC 0x4d0
+    /* Don't probe level-triggered interrupts -- reserved for PCI */
+    int level_mask = inb_p(PIC) | (inb_p(PIC+1) << 8);
+    if (level_mask)
+	mask0 &= ~level_mask;
+#endif
+    
     mask1 = mask2 = ni = 0;
     for (i = 0; i < 16; i++)
 	if ((mask0 & (1 << i)) && (try_irq(i) == 0))
@@ -251,6 +262,64 @@ static u_long irq_scan(u_long mask0)
     return (mask1 & mask2);
 }
 
+/*======================================================================
+
+    See if a card is present, powered up, in IO mode, and already
+    bound to a (non-PCMCIA) Linux driver.
+
+    We make an exception for cards that look like serial devices.
+    
+======================================================================*/
+
+static int is_active(int s)
+{
+    u_short scf1, ioctl, base, num;
+    u_char pwr, sstat;
+    u_long addr;
+    
+    tcic_setl(TCIC_ADDR, (s << TCIC_ADDR_SS_SHFT)
+	      | TCIC_ADDR_INDREG | TCIC_SCF1(s));
+    scf1 = tcic_getw(TCIC_DATA);
+    pwr = tcic_getb(TCIC_PWR);
+    sstat = tcic_getb(TCIC_SSTAT);
+    addr = TCIC_IWIN(s, 0);
+    tcic_setw(TCIC_ADDR, addr + TCIC_IBASE_X);
+    base = tcic_getw(TCIC_DATA);
+    tcic_setw(TCIC_ADDR, addr + TCIC_ICTL_X);
+    ioctl = tcic_getw(TCIC_DATA);
+
+    if (ioctl & TCIC_ICTL_TINY)
+	num = 1;
+    else {
+	num = (base ^ (base-1));
+	base = base & (base-1);
+    }
+
+    if ((sstat & TCIC_SSTAT_CD) && (pwr & TCIC_PWR_VCC(s)) &&
+	(scf1 & TCIC_SCF1_IOSTS) && (ioctl & TCIC_ICTL_ENA) &&
+	(check_region(base, num) != 0) && ((base & 0xfeef) != 0x02e8))
+	return 1;
+    else
+	return 0;
+}
+
+/*======================================================================
+
+    This returns the revision code for the specified socket.
+    
+======================================================================*/
+
+static int get_tcic_id(void)
+{
+    u_short id;
+    
+    tcic_aux_setw(TCIC_AUX_TEST, TCIC_TEST_DIAG);
+    id = tcic_aux_getw(TCIC_AUX_ILOCK);
+    id = (id & TCIC_ILOCKTEST_ID_MASK) >> TCIC_ILOCKTEST_ID_SH;
+    tcic_aux_setw(TCIC_AUX_TEST, 0);
+    return id;
+}
+
 /*====================================================================*/
 
 int tcic_init(void)
@@ -260,15 +329,11 @@ int tcic_init(void)
 
 #ifdef PCMCIA_DEBUG
     if (pc_debug)
-	printk(version);
+	printk(KERN_INFO "%s\n", version);
 #endif
     
-    printk("Databook TCIC-2 PCMCIA probe: ");
+    printk(KERN_INFO "Databook TCIC-2 PCMCIA probe: ");
     sock = 0;
-    if (ignore < 0) {
-	tcic_setb(TCIC_SCTRL, TCIC_SCTRL_RESET);
-	tcic_setb(TCIC_SCTRL, 0);
-    }
 
     if (check_region(tcic_base, 16) == 0) {
 	tcic_setw(TCIC_ADDR, 0);
@@ -276,6 +341,16 @@ int tcic_init(void)
 	    tcic_setw(TCIC_ADDR, 0xc3a5);
 	    if (tcic_getw(TCIC_ADDR) == 0xc3a5) sock = 2;
 	}
+	if (sock == 0) {
+	    /* See if resetting the controller does any good */
+	    tcic_setb(TCIC_SCTRL, TCIC_SCTRL_RESET);
+	    tcic_setb(TCIC_SCTRL, 0);
+	    tcic_setw(TCIC_ADDR, 0);
+	    if (tcic_getw(TCIC_ADDR) == 0) {
+		tcic_setw(TCIC_ADDR, 0xc3a5);
+		if (tcic_getw(TCIC_ADDR) == 0xc3a5) sock = 2;
+	    }
+	}
     }
     else
 	printk("could not allocate ports, ");
@@ -289,24 +364,42 @@ int tcic_init(void)
 
     sockets = 0;
     for (i = 0; i < sock; i++) {
-	if (i == ignore) continue;
+	if ((i == ignore) || is_active(i)) continue;
 	socket_table[sockets].psock = i;
 	socket_table[sockets].handler = NULL;
 	socket_table[sockets].info = NULL;
+	socket_table[sockets].id = get_tcic_id();
 	sockets++;
     }
 
+    switch (socket_table[0].id) {
+    case TCIC_ID_DB86082:
+	printk("DB86082"); break;
+    case TCIC_ID_DB86082A:
+	printk("DB86082A"); break;
+    case TCIC_ID_DB86084:
+	printk("DB86084"); break;
+    case TCIC_ID_DB86084A:
+	printk("DB86084A"); break;
+    case TCIC_ID_DB86072:
+	printk("DB86072"); break;
+    case TCIC_ID_DB86184:
+	printk("DB86184"); break;
+    case TCIC_ID_DB86082B:
+	printk("DB86082B"); break;
+    default:
+	printk("Unknown ID 0x%02x", socket_table[0].id);
+    }
+    
     /* Set up polling */
     poll_timer.function = &tcic_timer;
     poll_timer.data = 0;
     poll_timer.prev = poll_timer.next = NULL;
 
     /* Build interrupt mask */
-    printk("%d sockets\n  irq mask (", sockets);
-    mask = tcic_cap.irq_mask;
-    if (irq_mask) mask &= irq_mask;
-    scan = irq_scan(mask);
-    if (scan) {
+    printk(", %d sockets\n" KERN_INFO "  irq mask (", sockets);
+    mask = tcic_cap.irq_mask & irq_mask;
+    if (do_scan && (scan = irq_scan(mask))) {
 	mask = scan;
 	printk("scanned");
     }
@@ -315,16 +408,21 @@ int tcic_init(void)
     printk(") = 0x%lx, ", mask);
     tcic_cap.irq_mask = mask;
     
+    /* Check for only two interrupts available */
+    scan = (mask & (mask-1));
+    if (((scan & (scan-1)) == 0) && (poll_interval == 0))
+	poll_interval = HZ;
+    
     if (poll_interval == 0) {
 	if (cs_irq == 0) {
 	    for (i = 15; i > 0; i--)
 		if ((mask & (1 << i)) &&
-		    (request_irq(i, tcic_interrupt, 0, "tcic") == 0))
+		    (REQUEST_IRQ(i, tcic_interrupt, 0, "tcic") == 0))
 		    break;
 	    cs_irq = i;
 	}
 	else {
-	    if (request_irq(cs_irq, tcic_interrupt, 0, "tcic") != 0)
+	    if (REQUEST_IRQ(cs_irq, tcic_interrupt, 0, "tcic") == 0)
 		cs_irq = 0;
 	}
 	if (cs_irq == 0) poll_interval = 100;
@@ -343,13 +441,13 @@ int tcic_init(void)
     }
     
     /* jump start interrupt handler, if needed */
-    tcic_interrupt(0, NULL);
+    tcic_interrupt IRQ(0, NULL, NULL);
 
     if (register_ss_entry(sockets, &tcic_service) != 0) {
-	printk("tcic: register_ss_entry() failed\n");
+	printk(KERN_NOTICE "tcic: register_ss_entry() failed\n");
 	release_region(tcic_base, 16);
 	if (cs_irq != 0)
-	    free_irq(cs_irq);
+	    FREE_IRQ(cs_irq);
 	return -ENODEV;
     }
 
@@ -367,7 +465,7 @@ static void tcic_finish(void)
     cli();
     if (cs_irq != 0) {
 	tcic_aux_setw(TCIC_AUX_SYSCFG, TCIC_SYSCFG_AUTOBUSY|0x0a00);
-	free_irq(cs_irq);
+	FREE_IRQ(cs_irq);
     }
     if (timer_pending)
 	del_timer(&poll_timer);
@@ -377,7 +475,7 @@ static void tcic_finish(void)
 
 /*====================================================================*/
 
-static void tcic_interrupt(int irq, struct pt_regs *regs)
+static void tcic_interrupt IRQ(int irq, void *dev, struct pt_regs *regs)
 {
     int i, quick = 0;
     u_char latch, sstat;
@@ -386,7 +484,7 @@ static void tcic_interrupt(int irq, struct pt_regs *regs)
     static volatile int active = 0;
 
     if (active) {
-	printk("tcic: reentered interrupt handler!\n");
+	printk(KERN_NOTICE "tcic: reentered interrupt handler!\n");
 	return;
     }
     else
@@ -394,7 +492,7 @@ static void tcic_interrupt(int irq, struct pt_regs *regs)
 
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 1)
-	printk("tcic: tcic_interrupt()\n");
+	printk(KERN_DEBUG "tcic: tcic_interrupt()\n");
 #endif
     
     for (i = 0; i < sockets; i++) {
@@ -426,7 +524,7 @@ static void tcic_interrupt(int irq, struct pt_regs *regs)
 
     /* Schedule next poll, if needed */
     if (((cs_irq == 0) || quick) && (!timer_pending)) {
-	poll_timer.expires = quick ? poll_quick : poll_interval;
+	poll_timer.expires = RUN_AT(quick ? poll_quick : poll_interval);
 	add_timer(&poll_timer);
 	timer_pending = 1;
     }
@@ -434,7 +532,7 @@ static void tcic_interrupt(int irq, struct pt_regs *regs)
     
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 1)
-	printk("tcic: done\n");
+	printk(KERN_DEBUG "tcic: done\n");
 #endif
     
 } /* tcic_interrupt */
@@ -443,10 +541,10 @@ void tcic_timer(u_long data)
 {
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 1)
-	printk("tcic: tcic_timer()\n");
+	printk(KERN_DEBUG "tcic: tcic_timer()\n");
 #endif
     timer_pending = 0;
-    tcic_interrupt(0, NULL);
+    tcic_interrupt IRQ(0, NULL, NULL);
 } /* tcic_timer */
 
 /*====================================================================*/
@@ -455,14 +553,10 @@ static int tcic_register_callback(u_short lsock, ss_callback_t *call)
 {
     if (call == NULL) {
 	socket_table[lsock].handler = NULL;
-#ifdef MODULE
 	MOD_DEC_USE_COUNT;
-#endif
     }
     else {
-#ifdef MODULE
 	MOD_INC_USE_COUNT;
-#endif
 	socket_table[lsock].handler = call->handler;
 	socket_table[lsock].info = call->info;
     }
@@ -494,7 +588,8 @@ static int tcic_get_status(u_short lsock, u_int *value)
 	*value |= SS_POWERON;
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 1)
-	printk("tcic: GetStatus(%d) = %#2.2x\n", lsock, *value);
+	printk(KERN_DEBUG "tcic: GetStatus(%d) = %#2.2x\n",
+	       lsock, *value);
 #endif
     return 0;
 } /* tcic_get_status */
@@ -556,9 +651,10 @@ static int tcic_get_socket(u_short lsock, socket_state_t *state)
 
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 1)
-	printk("tcic: GetSocket(%d) = flags %#3.3lx, Vcc %d, Vpp %d"
-	       ", io_irq %d, csc_mask %#2.2lx\n", lsock, state->flags,
-	       state->Vcc, state->Vpp, state->io_irq, state->csc_mask);
+	printk(KERN_DEBUG "tcic: GetSocket(%d) = flags %#3.3lx, "
+	       "Vcc %d, Vpp %d, io_irq %d, csc_mask %#2.2lx\n", lsock,
+	       state->flags, state->Vcc, state->Vpp, state->io_irq,
+	       state->csc_mask);
 #endif
     return 0;
 } /* tcic_get_socket */
@@ -573,9 +669,10 @@ static int tcic_set_socket(u_short lsock, socket_state_t *state)
 
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 1)
-	printk("tcic: SetSocket(%d, flags %#3.3lx, Vcc %d, Vpp %d"
-	       ", io_irq %d, csc_mask %#2.2lx)\n", lsock, state->flags,
-	       state->Vcc, state->Vpp, state->io_irq, state->csc_mask);
+	printk(KERN_DEBUG "tcic: SetSocket(%d, flags %#3.3lx, "
+	       "Vcc %d, Vpp %d, io_irq %d, csc_mask %#2.2lx)\n",
+	       lsock, state->flags, state->Vcc, state->Vpp,
+	       state->io_irq, state->csc_mask);
 #endif
     tcic_setw(TCIC_ADDR+2, (psock << TCIC_SS_SHFT) | TCIC_ADR2_INDREG);
 
@@ -681,7 +778,7 @@ static int tcic_get_io_map(u_short lsock, struct pcmcia_io_map *io)
     }
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 1)
-	printk("tcic: GetIOMap(%d, %d) = %#2.2x, %d ns, "
+	printk(KERN_DEBUG "tcic: GetIOMap(%d, %d) = %#2.2x, %d ns, "
 	       "%#4.4x-%#4.4x\n", lsock, io->map, io->flags,
 	       io->speed, io->start, io->stop);
 #endif
@@ -698,7 +795,7 @@ static int tcic_set_io_map(u_short lsock, struct pcmcia_io_map *io)
     
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 1)
-	printk("tcic: SetIOMap(%d, %d, %#2.2x, %d ns, "
+	printk(KERN_DEBUG "tcic: SetIOMap(%d, %d, %#2.2x, %d ns, "
 	       "%#4.4x-%#4.4x)\n", lsock, io->map, io->flags,
 	       io->speed, io->start, io->stop);
 #endif
@@ -757,6 +854,7 @@ static int tcic_get_mem_map(u_short lsock, struct pcmcia_mem_map *mem)
     mem->flags = (mmap & TCIC_MMAP_REG) ? MAP_ATTRIB : 0;
     mmap &= TCIC_MMAP_CA_MASK;
     mem->card_start = mem->sys_start + (mmap << TCIC_MMAP_CA_SHFT);
+    mem->card_start &= 0x3ffffff;
     
     tcic_setw(TCIC_ADDR, addr + TCIC_MCTL_X);
     ctl = tcic_getw(TCIC_DATA);
@@ -767,9 +865,10 @@ static int tcic_get_mem_map(u_short lsock, struct pcmcia_mem_map *mem)
     
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 1)
-	printk("tcic: GetMemMap(%d, %d) = %#2.2x, %d ns, %#5.5lx-%#5.5lx,"
-	       " %#5.5lx\n", lsock, mem->map, mem->flags, mem->speed,
-	       mem->sys_start, mem->sys_stop, mem->card_start);
+	printk(KERN_DEBUG "tcic: GetMemMap(%d, %d) = %#2.2x, %d ns, "
+	       "%#5.5lx-%#5.5lx, %#5.5lx\n", lsock, mem->map,
+	       mem->flags, mem->speed, mem->sys_start, mem->sys_stop,
+	       mem->card_start);
 #endif
     return 0;
 } /* tcic_get_mem_map */
@@ -784,11 +883,12 @@ static int tcic_set_mem_map(u_short lsock, struct pcmcia_mem_map *mem)
 
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 1)
-	printk("tcic: SetMemMap(%d, %d, %#2.2x, %d ns, %#5.5lx-%#5.5lx,"
-	       " %#5.5lx)\n", lsock, mem->map, mem->flags, mem->speed,
-	       mem->sys_start, mem->sys_stop, mem->card_start);
+	printk(KERN_DEBUG "tcic: SetMemMap(%d, %d, %#2.2x, %d ns, "
+	       "%#5.5lx-%#5.5lx, %#5.5lx)\n", lsock, mem->map,
+	       mem->flags, mem->speed, mem->sys_start, mem->sys_stop,
+	       mem->card_start);
 #endif
-    if ((mem->map > 3) || (mem->card_start > 0xffffff) ||
+    if ((mem->map > 3) || (mem->card_start > 0x3ffffff) ||
 	(mem->sys_start > 0xffffff) || (mem->sys_stop > 0xffffff) ||
 	(mem->sys_start > mem->sys_stop) || (mem->speed > 1000))
 	return -EINVAL;
@@ -845,7 +945,7 @@ static int tcic_service(u_int lsock, u_int cmd, void *arg)
 
 #ifdef PCMCIA_DEBUG
     if (pc_debug) 
-	printk("tcic_ioctl(%d, %d, 0x%p)\n", lsock, cmd, arg);
+	printk(KERN_DEBUG "tcic_ioctl(%d, %d, 0x%p)\n", lsock, cmd, arg);
 #endif
 
     if (cmd < NFUNC)
@@ -858,20 +958,19 @@ static int tcic_service(u_int lsock, u_int cmd, void *arg)
 
 /*====================================================================*/
 
-#ifdef MODULE
-char kernel_version[] = UTS_RELEASE;
-
-int init_module(void)
+int init_tcic(void)
 {
     servinfo_t serv;
     CardServices(GetCardServicesInfo, &serv);
     if (serv.Revision != CS_RELEASE_CODE) {
-	printk("tcic: Card Services release does not match!\n");
+	printk(KERN_NOTICE "tcic: Card Services release "
+	       "does not match!\n");
 	return -1;
     }
     return tcic_init();
 }
 
+#ifdef MODULE
 void cleanup_module(void)
 {
     tcic_finish();
